# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Base delivery class."""

import copy
import socket
import logging
import smtplib

from lazr.config import as_boolean
from mailman.config import config
from mailman.interfaces.mta import IMailTransportAgentDelivery
from mailman.mta.connection import as_SecureMode, Connection
from public import public
from zope.interface import implementer


log = logging.getLogger('mailman.smtp')


@public
@implementer(IMailTransportAgentDelivery)
class BaseDelivery:
    """Base delivery class."""

    def __init__(self):
        """Create a basic deliverer."""
        self._connection = Connection(
            config.mta.smtp_host, int(config.mta.smtp_port),
            int(config.mta.max_sessions_per_connection),
            config.mta.smtp_user if config.mta.smtp_user else None,
            config.mta.smtp_pass if config.mta.smtp_pass else None,
            as_SecureMode(config.mta.smtp_secure_mode),
            as_boolean(config.mta.smtp_verify_cert),
            as_boolean(config.mta.smtp_verify_hostname),
            )

    def _deliver_to_recipients(self, mlist, msg, msgdata, recipients):
        """Low-level delivery to a set of recipients.

        :param mlist: The mailing list being delivered to.
        :type mlist: `IMailingList`
        :param msg: The original message being delivered.
        :type msg: `Message`
        :param msgdata: Additional message metadata for this delivery.
        :type msgdata: dictionary
        :param recipients: The recipients of this message.
        :type recipients: sequence
        :return: delivery failures as defined by `smtplib.SMTP.sendmail`
        :rtype: dictionary
        """
        # Do the actual sending.
        sender = self._get_sender(mlist, msg, msgdata)
        message_id = msg['message-id']
        # Since the recipients can be a set or a list, sort the recipients by
        # email address for predictability and testability.
        try:
            refused = self._connection.sendmail(
                sender, sorted(recipients), msg)
        except smtplib.SMTPRecipientsRefused as error:
            log.error('%s recipients refused: %s', message_id, error)
            refused = error.recipients
        except smtplib.SMTPResponseException as error:
            log.error('%s response exception: %s', message_id, error)
            refused = dict(
                # recipient -> (code, error)
                (recipient, (error.smtp_code, error.smtp_error))
                for recipient in recipients)
        except (socket.error, IOError, smtplib.SMTPException) as error:
            # MTA not responding, or other socket problems, or any other
            # kind of SMTPException.  In that case, nothing got delivered,
            # so treat this as a temporary failure.  We use error code 444
            # for this (temporary, unspecified failure, cf RFC 5321).
            log.error('%s low level smtp error: %s', message_id, error)
            error = str(error)
            refused = dict(
                # recipient -> (code, error)
                (recipient, (444, error))
                for recipient in recipients)
        return refused

    def _get_sender(self, mlist, msg, msgdata):
        """Return the envelope sender to use.

        The message metadata can override the calculation of the sender, but
        otherwise it falls to the list's -bounces robot.  If this message is
        not intended for any specific mailing list, the site owner's address
        is used.

        :param mlist: The mailing list being delivered to.
        :type mlist: `IMailingList`
        :param msg: The original message being delivered.
        :type msg: `Message`
        :param msgdata: Additional message metadata for this delivery.
        :type msgdata: dictionary
        :return: The envelope sender.
        :rtype: string
        """
        sender = msgdata.get('sender')
        if sender is None:
            return (config.mailman.site_owner
                    if mlist is None
                    else mlist.bounces_address)
        return sender


@public
class IndividualDelivery(BaseDelivery):
    """Deliver a unique individual message to each recipient.

    This is a framework delivery mechanism.  By using mixins, registration,
    and subclassing you can customize this delivery class to do any
    combination of VERP, full personalization, individualized header/footer
    decoration and even full mail merging.

    The core concept here is that for each recipient, the deliver() method
    iterates over the list of registered callbacks, each of which have a
    chance to modify the message before final delivery.
    """

    def __init__(self):
        """See `BaseDelivery`."""
        super().__init__()
        self.callbacks = []

    def deliver(self, mlist, msg, msgdata):
        """See `IMailTransportAgentDelivery`.

        Craft a unique message for every recipient.  Encode the recipient's
        delivery address in the return envelope so there can be no ambiguity
        in bounce processing.
        """
        refused = {}
        recipients = msgdata.get('recipients', set())
        for recipient in recipients:
            log.debug('IndividualDelivery to: %s', recipient)
            # Make a copy of the original messages and operator on it, since
            # we're going to munge it repeatedly for each recipient.
            message_copy = copy.deepcopy(msg)
            msgdata_copy = msgdata.copy()
            # Squirrel the current recipient away in the message metadata.
            # That way the subclass's _get_sender() override can encode the
            # recipient address in the sender, e.g. for VERP.
            msgdata_copy['recipient'] = recipient
            # See if the recipient is a member of the mailing list, and if so,
            # squirrel this information away for use by other modules, such as
            # the header/footer decorator.  XXX 2012-03-05 this is probably
            # highly inefficient on the database.
            member = mlist.members.get_member(recipient)
            msgdata_copy['member'] = member
            for callback in self.callbacks:
                callback(mlist, message_copy, msgdata_copy)
            status = self._deliver_to_recipients(
                mlist, message_copy, msgdata_copy, [recipient])
            refused.update(status)
        return refused



# Copyright (C) 2020-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""increase_info

Revision ID: ec5fe422e27c
Revises: 472e6d713047
Create Date: 2021-04-28 17:46:28.879085

"""

from alembic import op
from mailman.database.types import SAText, SAUnicode


# revision identifiers, used by Alembic.
revision = 'ec5fe422e27c'
down_revision = '472e6d713047'


def upgrade():
    # Mailing List info column needs to be bigger.
    with op.batch_alter_table('mailinglist') as batch_op:
        batch_op.alter_column('info', type_=SAText)


def downgrade():
    with op.batch_alter_table('mailinglist') as batch_op:
        batch_op.alter_column('info', type_=SAUnicode)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Style manager."""

from mailman.interfaces.configuration import ConfigurationUpdatedEvent
from mailman.interfaces.styles import (
    DuplicateStyleError,
    IStyle,
    IStyleManager,
)
from mailman.utilities.modules import add_components
from public import public
from zope.component import getUtility
from zope.interface import implementer
from zope.interface.verify import verifyObject


@public
@implementer(IStyleManager)
class StyleManager:
    """The built-in style manager."""

    def __init__(self):
        """Install all styles from the configuration files."""
        self._styles = {}

    def populate(self):
        self._styles.clear()
        add_components('styles', IStyle, self._styles)

    def get(self, name):
        """See `IStyleManager`."""
        return self._styles.get(name)

    @property
    def styles(self):
        """See `IStyleManager`."""
        for style_name in sorted(self._styles):
            yield self._styles[style_name]

    def register(self, style):
        """See `IStyleManager`."""
        verifyObject(IStyle, style)
        if style.name in self._styles:
            raise DuplicateStyleError(style.name)
        self._styles[style.name] = style

    def unregister(self, style):
        """See `IStyleManager`."""
        # Let KeyErrors percolate up.
        del self._styles[style.name]


@public
def handle_ConfigurationUpdatedEvent(event):
    if isinstance(event, ConfigurationUpdatedEvent):
        getUtility(IStyleManager).populate()



# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Virgin runner.

This runner handles messages that the Mailman system gives virgin birth to.
E.g. acknowledgment responses to user posts or replybot messages.  They need
to go through some minimal processing before they can be sent out to the
recipient.
"""

from mailman.core.pipelines import process
from mailman.core.runner import Runner
from public import public


@public
class VirginRunner(Runner):
    def _dispose(self, mlist, msg, msgdata):
        # We need to fast track this message through any pipeline handlers
        # that touch it, e.g. especially cook-headers.
        msgdata['_fasttrack'] = True
        # Use the 'virgin' pipeline.
        process(mlist, msg, msgdata, 'virgin')
        # Do not keep this message queued.
        return False



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for mailman.utilities.modules."""

import os
import sys
import unittest

from contextlib import contextmanager, ExitStack
from mailman.interfaces.rules import IRule
from mailman.interfaces.styles import IStyle
from mailman.testing.helpers import configuration
from mailman.testing.layers import ConfigLayer
from mailman.utilities.filesystem import path
from mailman.utilities.modules import (
    find_components,
    find_pluggable_components,
    hacked_sys_modules,
)
from pathlib import Path
from tempfile import TemporaryDirectory


@contextmanager
def hack_syspath(index, path):
    old_path = sys.path[:]
    try:
        sys.path.insert(index, path)
        yield
    finally:
        sys.path = old_path


def clean_mypackage():
    # Make a copy since we'll mutate it as we go.
    for module in sys.modules.copy():
        package, dot, rest = module.partition('.')
        if package == 'mypackage':
            del sys.modules[module]


class TestModuleImports(unittest.TestCase):
    layer = ConfigLayer

    def test_find_modules_with_dotfiles(self):
        # Emacs creates lock files when a single file is opened by more than
        # one user. These files look like .#<filename>.py because of which
        # find_components() tries to import them but fails. All such files
        # should be ignored by default.
        with ExitStack() as resources:
            # Creating a temporary directory and adding it to sys.path.
            temp_package = resources.enter_context(TemporaryDirectory())
            resources.enter_context(hack_syspath(0, temp_package))
            resources.callback(clean_mypackage)
            # Create a module inside the above package along with a good, bad
            # and __init__.py file so that we can import from it.
            module_path = os.path.join(temp_package, 'mypackage')
            os.mkdir(module_path)
            init_file = os.path.join(module_path, '__init__.py')
            Path(init_file).touch()
            good_file = os.path.join(module_path, 'goodfile.py')
            bad_file = os.path.join(module_path, '.#badfile.py')
            with open(good_file, 'w', encoding='utf-8') as fp:
                print("""\
from public import public
from mailman.interfaces.styles import IStyle
from zope.interface import implementer

@public
@implementer(IStyle)
class GoodStyle:
    name = 'good-style'
    def apply(self):
        pass
""", file=fp)
            with open(bad_file, 'w', encoding='utf-8') as fp:
                print("""\
from public import public
from mailman.interfaces.styles import IStyle
from zope.interface import implementer

@public
@implementer(IStyle)
class BadStyle:
    name = 'bad-style'
    def apply(self):
        pass
""", file=fp)
            # Find all the IStyle components in the dummy package.  This
            # should find GoodStyle but not BadStyle.
            names = [component.name
                     for component
                     in find_components('mypackage', IStyle)]
            self.assertEqual(names, ['good-style'])

    def test_find_components_abstract_component(self):
        # find_components() finds the class unless it's been
        # decorated with the @abstract_component decorator.
        with ExitStack() as resources:
            # Creating a temporary directory and adding it to sys.path.
            temp_package = resources.enter_context(TemporaryDirectory())
            resources.enter_context(hack_syspath(0, temp_package))
            resources.callback(clean_mypackage)
            # Create a module inside the above package along with an
            # __init__.py file so that we can import from it.
            module_path = os.path.join(temp_package, 'mypackage')
            os.mkdir(module_path)
            init_file = os.path.join(module_path, '__init__.py')
            Path(init_file).touch()
            component_file = os.path.join(module_path, 'components.py')
            with open(component_file, 'w', encoding='utf-8') as fp:
                print("""\
from mailman.interfaces.styles import IStyle
from mailman.utilities.modules import abstract_component
from public import public
from zope.interface import implementer

@public
@implementer(IStyle)
class ConcreteStyle:
    name = 'concrete-style'
    def apply(self):
        pass

@public
@implementer(IStyle)
@abstract_component
class AbstractStyle:
    name = 'abstract-style'
    def apply(self):
        pass
""", file=fp)
            names = [component.name
                     for component
                     in find_components('mypackage', IStyle)]
            self.assertEqual(names, ['concrete-style'])

    def test_hacked_sys_modules(self):
        self.assertIsNone(sys.modules.get('mailman.not_a_module'))
        with hacked_sys_modules('mailman.not_a_module', object()):
            self.assertIsNotNone(sys.modules.get('mailman.not_a_module'))

    def test_hacked_sys_modules_restore(self):
        email_package = sys.modules['email']
        sentinel = object()
        with hacked_sys_modules('email', sentinel):
            self.assertEqual(sys.modules.get('email'), sentinel)
        self.assertEqual(sys.modules.get('email'), email_package)

    def test_find_pluggable_components_by_plugin_name(self):
        with ExitStack() as resources:
            testing_path = resources.enter_context(
                path('mailman.plugins.testing',  ''))
            resources.enter_context(hack_syspath(0, str(testing_path)))
            resources.enter_context(configuration('plugin.example', **{
                'class': 'example.hooks.ExamplePlugin',
                'enabled': 'yes',
                }))
            components = list(find_pluggable_components('rules', IRule))
        self.assertIn('example-rule', {rule.name for rule in components})

    def test_find_pluggable_components_by_component_package(self):
        with ExitStack() as resources:
            testing_path = resources.enter_context(
                path('mailman.plugins.testing', ''))
            resources.enter_context(hack_syspath(0, str(testing_path)))
            resources.enter_context(configuration('plugin.example', **{
                'class': 'example.hooks.ExamplePlugin',
                'enabled': 'yes',
                'component_package': 'alternate',
                }))
            components = list(find_pluggable_components('rules', IRule))
        self.assertNotIn('example-rule', {rule.name for rule in components})
        self.assertIn('alternate-rule', {rule.name for rule in components})



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test some plugin behavior."""

import unittest

from contextlib import ExitStack
from mailman.testing.documentation import run_mailman
from mailman.testing.layers import ConfigLayer
from tempfile import NamedTemporaryFile


class TestExternalHooks(unittest.TestCase):
    layer = ConfigLayer
    maxDiff = None

    def setUp(self):
        self.resources = ExitStack()
        self.addCleanup(self.resources.close)
        self.config_file = self.resources.enter_context(NamedTemporaryFile())

    def test_pre_hook_deprecated(self):
        with open(self.config_file.name, 'w', encoding='utf-8') as fp:
            print("""\
[mailman]
pre_hook: sys.exit

[logging.plugins]
propagate: yes
""", file=fp)
        proc = run_mailman(['-C', self.config_file.name, 'info'])
        # We only care about the log warning printed to stdout.
        warning = proc.stdout.splitlines()[0]
        self.assertEqual(
            warning[-111:],
            'The [mailman]pre_hook configuration value has been replaced '
            "by the plugins infrastructure, and won't be called.")

    def test_post_hook_deprecated(self):
        with open(self.config_file.name, 'w', encoding='utf-8') as fp:
            print("""\
[mailman]
post_hook: sys.exit

[logging.plugins]
propagate: yes
""", file=fp)
        proc = run_mailman(['-C', self.config_file.name, 'info'])
        # We only care about the log warning printed to stdout.
        warning = proc.stdout.splitlines()[0]
        self.assertEqual(
            warning[-112:],
            'The [mailman]post_hook configuration value has been replaced '
            "by the plugins infrastructure, and won't be called.")



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Web service helpers."""

import json
import types
import falcon
import hashlib

from contextlib import suppress
from datetime import datetime, timedelta
from email.header import Header
from email.message import Message
from enum import Enum
from functools import partial
from lazr.config import as_boolean
from mailman.config import config
from pprint import pformat
from public import public


CONTENT_TYPE_JSON = 'application/json; charset=UTF-8'
CONTENT_TYPE_TEXT_PLAIN = 'text/plain'


class ExtendedEncoder(json.JSONEncoder):
    """An extended JSON encoder which knows about other data types."""

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            # as_timedelta() does not recognize microseconds, so convert these
            # to floating seconds, but only if there are any seconds.
            if obj.seconds > 0 or obj.microseconds > 0:
                seconds = obj.seconds + obj.microseconds / 1000000.0
                return '{}d{}s'.format(obj.days, seconds)
            return '{}d'.format(obj.days)
        elif isinstance(obj, Enum):
            # It's up to the decoding validator to associate this name with
            # the right Enum class.
            return obj.name
        elif isinstance(obj, bytes):
            return bytes_to_str(obj)
        elif isinstance(obj, Message):
            return obj.as_string()
        elif isinstance(obj, Header):
            return str(obj)
        return super().default(obj)


def bytes_to_str(value):
    # Convert a string to unicode when the encoding is not declared.
    if not isinstance(value, bytes):
        return value
    for encoding in ('ascii', 'utf-8', 'raw_unicode_escape'):
        with suppress(UnicodeDecodeError):
            return value.decode(encoding)


@public
def etag(resource):
    """Calculate the etag and return a JSON representation.

    The input is a dictionary representing the resource.  This
    dictionary must not contain an `http_etag` key.  This function
    calculates the etag by using the sha1 hexdigest of the
    pretty-printed (and thus key-sorted and predictable) representation
    of the dictionary.  It then inserts this value under the `http_etag`
    key, and returns the JSON representation of the modified dictionary.

    :param resource: The original resource representation.
    :type resource: dictionary
    :return: JSON representation of the modified dictionary.
    :rtype string
    """
    assert 'http_etag' not in resource, 'Resource already etagged'
    # Calculate the tag from a predictable (i.e. sorted) representation of the
    # dictionary.  The actual details aren't so important.  pformat() is
    # guaranteed to sort the keys, however it returns a str and the hash
    # library requires a bytes.  Use the safest possible encoding.
    hashfood = pformat(resource).encode('raw-unicode-escape')
    etag = hashlib.sha1(hashfood).hexdigest()
    resource['http_etag'] = '"{}"'.format(etag)
    return json.dumps(resource, cls=ExtendedEncoder,
                      sort_keys=as_boolean(config.devmode.enabled))


@public
class CollectionMixin:
    """Mixin class for common collection-ish things."""

    def _resource_as_dict(self, resource, fields=None):
        """Return the dictionary representation of a resource.

        This must be implemented by subclasses.

        :param resource: The resource object.
        :type resource: object
        :param fields: The resource fields which should be included.
        :type fields: List[str]
        :return: The representation of the resource.
        :rtype: dict
        """
        raise NotImplementedError

    def _resource_as_json(self, resource):
        """Return the JSON formatted representation of the resource."""
        resource = self._resource_as_dict(resource)
        assert resource is not None, resource
        return etag(resource)

    def _get_collection(self, request):
        """Return the collection as a sequence.

        The returned value must support the collections.abc.Sequence
        API.  This method must be implemented by subclasses.

        :param request: An http request.
        :return: The collection
        :rtype: collections.abc.Sequence
        """
        raise NotImplementedError

    def _paginate(self, request, collection):
        """Method to paginate through collection result lists.

        Use this to return only a slice of a collection, specified in
        the request itself.  The request should use query parameters
        `count` and `page` to specify the slice they want.  The slice
        will start at index ``(page - 1) * count`` and end (exclusive)
        at ``(page * count)``.
        """
        # Allow falcon's HTTPBadRequest exceptions to percolate up.  They'll
        # get turned into HTTP 400 errors.
        count = request.get_param_as_int('count')
        page = request.get_param_as_int('page')
        total_size = len(collection)
        if count is None and page is None:
            return 0, total_size, collection
        # TODO(maxking): Count and page should be positive integers. Once
        # falcon 2.0.0 is out and we can jump to it, we can remove this logic
        # and use `min_value` parameter in request.get_param_as_int.
        if count < 0:
            raise falcon.HTTPInvalidParam(
                count, 'count should be a positive integer.')
        if page < 1:
            raise falcon.HTTPInvalidParam(
                page, 'page should be greater than 0.')
        list_start = (page - 1) * count
        list_end = page * count
        return list_start, total_size, collection[list_start:list_end]

    def _make_collection(self, request, fields=None):
        """Provide the collection to the REST layer."""
        start, total_size, collection = self._paginate(
            request, self._get_collection(request))
        result = dict(start=start, total_size=total_size)
        if len(collection) != 0:
            # XXX(maxking): This is not the nicest way to use the mixin class,
            # but this is just meant to minimize the code changes since all the
            # resource endpoints do not support fields API. We pass on the
            # fields to underlying serialization method conditionally.
            as_dict = self._resource_as_dict
            if fields is not None:
                as_dict = partial(self._resource_as_dict, fields=fields)  # pragma: nocover  # noqa: E501
            entries = [as_dict(resource) for resource in collection]
            assert None not in entries, entries
            # Create the collection resource
            [etag(resource) for resource in entries]
            result['entries'] = entries
        return result


@public
class GetterSetter:
    """Get and set attributes on an object.

    Most attributes are fairly simple - a getattr() or setattr() on the object
    does the trick, with the appropriate encoding or decoding on the way in
    and out.  Encoding doesn't happen here though; the standard JSON library
    handles most types, but see ExtendedEncoder for additional support.

    Others are more complicated since they aren't kept in the model as direct
    columns in the database.  These will use subclasses of this base class.
    Read-only attributes will have a decoder which always raises ValueError.
    """

    def __init__(self, decoder=None):
        """Create a getter/setter for a specific attribute.

        :param decoder: The callable for decoding a web request value string
            into the specific data type needed by the object's attribute.  Use
            None to indicate a read-only attribute.  The callable should raise
            ValueError when the web request value cannot be converted.
        :type decoder: callable
        """
        self.decoder = decoder

    def get(self, obj, attribute):
        """Return the named object attribute value.

        :param obj: The object to access.
        :type obj: object
        :param attribute: The attribute name.
        :type attribute: string
        :return: The attribute value, ready for JSON encoding.
        :rtype: object
        """
        value = getattr(obj, attribute)
        # If the attribute is a generator type, return a list instead.
        if isinstance(value, types.GeneratorType):
            value = list(value)
        return value

    def put(self, obj, attribute, value):
        """Set the named object attribute value.

        :param obj: The object to change.
        :type obj: object
        :param attribute: The attribute name.
        :type attribute: string
        :param value: The new value for the attribute.
        """
        setattr(obj, attribute, value)

    def __call__(self, value):
        """Convert the value to its internal format.

        :param value: The web request value to convert.
        :type value: string
        :return: The converted value.
        :rtype: object
        """
        if self.decoder is None:
            return value
        if self.decoder is as_boolean and isinstance(value, bool):
            return value
        return self.decoder(value)


# Falcon REST framework add-ons.

@public
def child(matcher=None):
    def decorator(func):
        if matcher is None:
            func.__matcher__ = func.__name__
        else:
            func.__matcher__ = matcher
        return func
    return decorator


@public
class ChildError:
    def __init__(self, status):
        self._status = status

    def _oops(self, request, response):
        raise falcon.HTTPError(self._status, title=None)  # pragma: nocover

    on_get = _oops
    on_post = _oops
    on_put = _oops
    on_patch = _oops
    on_delete = _oops


@public
class BadRequest(ChildError):
    def __init__(self):
        super().__init__(falcon.HTTP_400)


@public
class NotFound(ChildError):
    def __init__(self):
        super().__init__(falcon.HTTP_404)


@public
def okay(response, body=None):
    response.status = falcon.HTTP_200
    if body is not None:
        response.text = body


@public
def no_content(response):
    response.status = falcon.HTTP_204


@public
def not_found(response, body='404 Not Found'):
    response.status = falcon.HTTP_404
    response.content_type = CONTENT_TYPE_JSON
    if isinstance(body, bytes):
        body = body.decode()
    if body is not None:
        response.text = falcon.HTTPNotFound(description=body).to_json()


@public
def accepted(response, body=None):
    response.status = falcon.HTTP_202
    if body is not None:
        response.text = body


@public
def bad_request(response, body='400 Bad Request'):
    response.status = falcon.HTTP_400
    response.content_type = CONTENT_TYPE_JSON
    if isinstance(body, bytes):
        body = body.decode()
    if body is not None:
        response.text = falcon.HTTPBadRequest(description=body).to_json()


@public
def created(response, location):
    response.status = falcon.HTTP_201
    response.location = location


@public
def conflict(response, body='409 Conflict'):
    response.status = falcon.HTTP_409
    response.content_type = CONTENT_TYPE_JSON
    if isinstance(body, bytes):
        body = body.decode()
    if body is not None:
        response.text = falcon.HTTPConflict(description=body).to_json()


@public
def forbidden(response, body='403 Forbidden'):
    response.status = falcon.HTTP_403
    response.content_type = CONTENT_TYPE_JSON
    if isinstance(body, bytes):
        body = body.decode()
    if body is not None:
        response.text = falcon.HTTPForbidden(description=body).to_json()


@public
def get_request_params(request):
    """Return the request items based on the content-type header"""
    # maxking: If there is a requirement for a new media type in future, The
    # way to handle that would be add a new media type handler in Falcon's API
    # class and then use the items returned by that handler here to return the
    # values to validators in the form of a dictionary.

    # We parse the request based on the content type. Falcon has a default
    # JSONHandler handler to parse json media type, so we can just do
    # `request.media` to return the request params passed as json body.
    if (request.content_type and
            request.content_type.startswith('application/json')):
        return request.media or dict()
    # request.params returns the parameters passed as URL form encoded.
    return request.params or dict()



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `queues` resource."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.database.transaction import transaction
from mailman.testing.helpers import call_api, get_queue_messages
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError


TEXT = """\
From: anne@example.com
To: test@example.com
Subject: A test
Message-ID: <ant>

"""


class TestQueues(unittest.TestCase):
    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('test@example.com')

    def test_missing_queue(self):
        # Trying to print a missing queue gives a 404.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/queues/notaq')
        self.assertEqual(cm.exception.code, 404)

    def test_no_such_list(self):
        # POSTing to a queue with a bad list-id gives a 400.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/queues/bad', {
                'list_id': 'nosuchlist.example.com',
                'text': TEXT,
                })
        self.assertEqual(cm.exception.code, 400)

    def test_inject(self):
        # Injecting a message leaves the message in the queue.
        starting_messages = get_queue_messages('bad')
        self.assertEqual(len(starting_messages), 0)
        content, response = call_api('http://localhost:9001/3.0/queues/bad', {
            'list_id': 'test.example.com',
            'text': TEXT})
        self.assertEqual(response.status_code, 201)
        location = response.headers['location']
        filebase = location.split('/')[-1]
        # The message is in the 'bad' queue.
        content, response = call_api('http://localhost:9001/3.0/queues/bad')
        files = content['files']
        self.assertEqual(len(files), 1)
        self.assertEqual(files[0], filebase)
        # Verify the files directly.
        files = list(config.switchboards['bad'].files)
        self.assertEqual(len(files), 1)
        self.assertEqual(files[0], filebase)
        # Verify the content.
        items = get_queue_messages('bad')
        self.assertEqual(len(items), 1)
        msg = items[0].msg
        # Remove some headers that get added by Mailman.
        del msg['date']
        self.assertEqual(msg['message-id-hash'],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')
        del msg['message-id-hash']
        del msg['x-message-id-hash']
        self.assertMultiLineEqual(msg.as_string(), TEXT)

    def test_delete_file(self):
        # Inject a file, then delete it.
        content, response = call_api('http://localhost:9001/3.0/queues/bad', {
            'list_id': 'test.example.com',
            'text': TEXT})
        location = response.headers['location']
        self.assertEqual(len(config.switchboards['bad'].files), 1)
        # Delete the file through REST.
        content, response = call_api(location, method='DELETE')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(len(config.switchboards['bad'].files), 0)



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Testing functions in the filesystem utilities."""

import os
import shutil
import tempfile
import unittest

from mailman.utilities.filesystem import first_inexistent_directory, makedirs


def fake_makedirs(path, mode):
    """A fake makedirs function"""

    with open(path, 'a'):
        pass

    raise FileExistsError("%s exists.", path)


class TestMakedirs(unittest.TestCase):
    """Tests the makedirs utility function"""

    def setUp(self):
        """Makes a temp directory and defines some paths to create. Then, check
        what's the return of the functions."""

        self.test_directory = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.test_directory)

        self.foo = os.path.join(self.test_directory, "foo")
        self.bar = os.path.join(self.foo, "bar")
        self.baz = os.path.join(self.bar, "baz")

    def test_first_inexistent_directory(self):
        """Tests the output of first_inexistent_directory"""

        upwards_dir = first_inexistent_directory(self.baz)

        self.assertEqual(upwards_dir, self.foo)

    def test_first_inexistent_directory_traversing_nondir(self):
        """Tests the output of first_inexistent_directory when one of the
        intermediate directories is actually a file."""

        os.makedirs(self.foo, 0o0755)
        with open(self.bar, 'a'):
            pass

        with self.assertRaises(FileExistsError):
            first_inexistent_directory(self.baz)

    def test_makedirs_race_condition(self):
        """Mocks os.makedirs behaviour to create willingly a race condition in
        filesystem.makedirs and test it."""

        first_inexistent_directory(self.baz)

        makedirs(self.bar)

        with unittest.mock.patch('os.makedirs', new=fake_makedirs):
            with self.assertRaises(FileExistsError):
                makedirs(self.baz)



# Copyright (C) 2020-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.
"""add_archive_rendering_mode

Revision ID: 9735f5e5dbdb
Revises: d151c0b8d6f7
Create Date: 2020-05-14 15:59:26.277647

"""
import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = '9735f5e5dbdb'
down_revision = 'd151c0b8d6f7'


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mailinglist', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column('archive_rendering_mode', sa.Integer(), nullable=True))
    # Default it to 'text' for all the MailingList since this is a new field.
    op.execute("UPDATE mailinglist SET archive_rendering_mode = 1")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mailinglist', schema=None) as batch_op:
        batch_op.drop_column('archive_rendering_mode')
    # ### end Alembic commands ###



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The email command 'help'."""

from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from mailman.utilities.string import wrap
from public import public
from zope.interface import implementer


SPACE = ' '


@public
@implementer(IEmailCommand)
class Help:
    """The email 'help' command."""

    name = 'help'
    argument_description = '[command]'
    description = _('Get help about available email commands.')
    short_description = description

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        # With no argument, print the command and a short description, which
        # is contained in the short_description attribute.
        if len(arguments) == 0:
            length = max(len(command) for command in config.commands)
            format = '{{0: <{0}s}} - {{1}}'.format(length)
            for command_name in sorted(config.commands):
                command = config.commands[command_name]
                short_description = getattr(
                    command, 'short_description', _('n/a'))
                print(format.format(command.name, short_description),
                      file=results)
            return ContinueProcessing.yes
        elif len(arguments) == 1:
            command_name = arguments[0]
            command = config.commands.get(command_name)
            if command is None:
                print(_('${self.name}: no such command: ${command_name}'),
                      file=results)
                return ContinueProcessing.no
            print('{} {}'.format(command.name, command.argument_description),
                  file=results)
            print(command.short_description, file=results)
            if command.short_description != command.description:
                print(wrap(command.description), file=results)
            return ContinueProcessing.yes
        else:
            printable_arguments = SPACE.join(arguments)           # noqa: F841
            print(_('${self.name}: too many arguments: '
                    '${printable_arguments}'),
                  file=results)
            return ContinueProcessing.no



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Internationalization for the tests."""

from flufl.i18n import registry
from gettext import GNUTranslations, NullTranslations
from importlib_resources import open_binary
from public import public


class TestingStrategy:
    """A strategy that finds catalogs in the testing directory."""

    def __init__(self, name):
        self.name = name

    def __call__(self, language_code=None):
        if language_code in ('en', None):
            return NullTranslations()
        mo_file = 'mailman-%s.mo' % language_code
        with open_binary('mailman.testing', mo_file) as fp:
            return GNUTranslations(fp)


@public
def initialize():
    """Install a global underscore function for testing purposes."""
    # Avoid circular imports.
    from mailman.core.i18n import initialize as core_initialize
    strategy = TestingStrategy('mailman-testing')
    application = registry.register(strategy)
    core_initialize(application)



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Bounce support."""

import logging
import datetime

from email.utils import make_msgid
from lazr.config import as_boolean, as_timedelta
from mailman.app.bounces import _ProbePendable, PENDABLE_LIFETIME, send_probe
from mailman.app.membership import delete_member
from mailman.app.notifications import (
    send_admin_disable_notice,
    send_admin_increment_notice,
    send_admin_removal_notice,
    send_user_disable_warning,
)
from mailman.config import config
from mailman.database.model import Model
from mailman.database.transaction import (
    dbconnection,
    transaction,
    transactional,
)
from mailman.database.types import Enum, SAUnicode
from mailman.interfaces.bounce import (
    BounceContext,
    IBounceEvent,
    IBounceProcessor,
    InvalidBounceEvent,
)
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import DeliveryStatus, IMembershipManager
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.pending import IPendings
from mailman.utilities.datetime import now
from public import public
from sqlalchemy import Boolean, Column, DateTime, Integer
from zope.component import getUtility
from zope.interface import implementer


log = logging.getLogger('mailman.bounce')


@public
@implementer(IBounceEvent)
class BounceEvent(Model):
    """See `IBounceEvent`."""

    __tablename__ = 'bounceevent'

    id = Column(Integer, primary_key=True)
    list_id = Column(SAUnicode)
    email = Column(SAUnicode)
    timestamp = Column(DateTime)
    message_id = Column(SAUnicode)
    context = Column(Enum(BounceContext))
    processed = Column(Boolean)

    def __init__(self, list_id, email, msg, context=None):
        self.list_id = list_id
        self.email = email
        self.timestamp = now()
        msgid = msg['message-id']
        self.message_id = msgid
        self.context = (BounceContext.normal if context is None else context)
        self.processed = False


@public
@implementer(IBounceProcessor)
class BounceProcessor:
    """See `IBounceProcessor`."""

    @dbconnection
    def register(self, store, mlist, email, msg, where=None):
        """See `IBounceProcessor`."""
        # Save the DSN in the message store for notices.  It doesn't
        # matter if we save it more than once.  Only one copy will be
        # saved, but ensure it has a Message-ID so we can retreive it.
        if msg.get('message-id') is None:
            msg['Message-ID'] = make_msgid          # pragma: nocover
        getUtility(IMessageStore).add(msg)
        # We also need to pend a token for this or the message will be
        # removed as an orphan by the task runner.  We don't need much
        # from this.  We pend the msgid as _mod_message_id for the
        # task runner.
        pendable = _ProbePendable(
            _mod_message_id=msg.get('message-id'))
        getUtility(IPendings).add(
            pendable, lifetime=as_timedelta(PENDABLE_LIFETIME))
        event = BounceEvent(mlist.list_id, email, msg, where)
        store.add(event)
        return event

    @property
    @dbconnection
    def events(self, store):
        """See `IBounceProcessor`."""
        yield from store.query(BounceEvent).all()

    @property
    @dbconnection
    def unprocessed(self, store):
        """See `IBounceProcessor`."""
        yield from store.query(BounceEvent).filter_by(processed=False)

    def _disable_delivery(self, mlist, member, event):
        """Disable deliver for the member and maybe notify the admin.

        :param mlist: The mailing list which bounce event belongs to.
        :type mlist: IMailingList
        :param member: The member object the bouncing address belongs to.
        :type member: IMember
        :param event: The bounce event that causes this.
        :type event:" IBounceEvent
        """
        # If the membership is already disabled, do not sent another notice to
        # the admin.
        if member.preferences.delivery_status == DeliveryStatus.by_bounces:
            return
        member.preferences.delivery_status = DeliveryStatus.by_bounces
        # We also need to set these.  It doesn't matter if they are already
        # set from a prior disable.  They should be set no for this one.
        member.total_warnings_sent = 0
        member.last_warning_sent = datetime.datetime.min
        log.info('Disabling delivery for %s on list %s by bounce',
                 event.email, mlist.list_id)
        if mlist.bounce_notify_owner_on_disable:
            send_admin_disable_notice(
                mlist, event, display_name=member.display_name)

    @transactional
    @dbconnection
    def process_event(self, store, event):
        """See `IBounceProcessor`."""
        list_manager = getUtility(IListManager)
        mlist = list_manager.get(event.list_id)
        if mlist is None:
            # List was removed before the bounce is processed.
            event.processed = True
            # This needs an explicit commit because of the raise.
            config.db.commit()
            raise InvalidBounceEvent(
                'Bounce for non-existent list {}'.format(event.list_id))
        member = mlist.members.get_member(event.email)
        if member is None:
            event.processed = True
            # This needs an explicit commit because of the raise.
            config.db.commit()
            raise InvalidBounceEvent(
                'Email {} is not a subcriber of {}'.format(
                    event.email, mlist.list_id))

        # If this is a probe bounce, that we are sent before to check for this
        # Mailbox, we just disable the delivery for this member.
        if event.context == BounceContext.probe:
            log.info('Probe bounce received for member %s on list %s.',
                     event.email, mlist.list_id)
            event.processed = True
            self._disable_delivery(mlist, member, event)
            return

        # Looks like this is a regular bounce event, we need to process it
        # in the follow order:
        # 0. If the member is already disabled by bounce, we ignore this
        #    event.
        # 1. If the date of the bounce is same as the previous bounce, we
        #    ignore this event.
        # 2. Check if the bounce info is valid, bump the bounce_score and
        #    update the last bounce info.
        # 3. If the bounce info is stale, reset the bounce score with
        #    this new value.
        # 4. If the bounce_score is greater than threshold after the above,
        #    a) Send a VERP probe, if configured to do so
        #    b) Disable membership otherwise and notify the user and
        #       warnings.
        if member.preferences.delivery_status == DeliveryStatus.by_bounces:
            log.info('Residual bounce received for member %s on list %s.',
                     event.email, mlist.list_id)
            event.processed = True
            return
        if (member.last_bounce_received is not None and
                member.last_bounce_received.date() == event.timestamp.date()):
            # Update the timestamp of the last bounce received.
            member.last_bounce_received = event.timestamp
            event.processed = True
            log.info('Member %s already scored a bounce on list %s today.',
                     event.email, mlist.list_id)
            return

        if member.last_bounce_received is None or (
                member.last_bounce_received <
                event.timestamp - mlist.bounce_info_stale_after):
            # Reset the bounce score to 1, for the current bounce that we got.
            member.bounce_score = 1
        else:
            # Update the bounce score to reflect this bounce.
            member.bounce_score += 1
        # Update the last received time for the bounce.
        member.last_bounce_received = event.timestamp
        log.info('Member %s on list %s, bounce score = %d.', event.email,
                 mlist.list_id, member.bounce_score)
        # Now, we are done updating. Let's see if the threshold is reached and
        # disable based on that, but first if we're not going to disable we
        # need to send the bounce score incremented notice if requested.
        if mlist.bounce_notify_owner_on_bounce_increment and (
                member.bounce_score < mlist.bounce_score_threshold or
                not as_boolean(config.mta.verp_probes)):
            send_admin_increment_notice(
                mlist, event, display_name=member.display_name)
        if member.bounce_score >= mlist.bounce_score_threshold:
            # Save bounce_score because sending probe resets it.
            saved_bounce_score = member.bounce_score
            # Try to get the dsn from the message store.  It should be there.
            msg = getUtility(IMessageStore).get_message_by_id(event.message_id)
            if as_boolean(config.mta.verp_probes):
                send_probe(member, msg=msg, message_id=event.message_id)
                action = 'sending probe'
            else:
                self._disable_delivery(mlist, member, event)
                action = 'disabling delivery'
            log.info(
                'Member %s on list %s, bounce score %d >= threshold %d, %s.',
                event.email, mlist.list_id, saved_bounce_score,
                mlist.bounce_score_threshold, action)
        event.processed = True

    @transactional
    @dbconnection
    def send_warnings_and_remove(self, store):
        """Send a warning email to the users who are disabled, if needed.

        Also, if the max number of warnings have already been sent, remove the
        member.
        """
        # Query the database for all the Members, who have:

        # 1. Last warning was sent more than Mailinglist's disable notice
        #    warnings interval ago.
        # 2. Total warnings sent are less that Mailinglist's maximum number of
        #    warnings to be sent before the member is removed.
        self._send_warnings()
        self._remove_memberships()

    @dbconnection
    def _remove_memberships(self, store):
        """Remove all the memberships for whom max number of warnings have
        been sent.
        """
        manager = getUtility(IMembershipManager)
        for member in manager.memberships_pending_removal():
            # We do not want to send duplicate notifications to the
            # administrators, so we send them a membership change event only if
            # the list is configured not to notify them about removal due to
            # bounces.
            # Although, note that if admin_notif is None, they will receive
            # notification only if the mailinglist is configured to notify on
            # membership changes, which is a different setting..
            admin_notif = None
            send_admin_notif = False
            if member.mailing_list.bounce_notify_owner_on_removal:
                admin_notif = False
                send_admin_notif = True

            with transaction():
                delete_member(
                    mlist=member._mailing_list, email=member.address.email,
                    admin_notif=admin_notif, userack=True)

            if send_admin_notif:
                send_admin_removal_notice(
                    member.mailing_list, member.address.email,
                    member.display_name)

            log.info('Removed %s as a member of %s mailing list due to '
                     'excessive bounces', member.address.email,
                     member._mailing_list.display_name)

    @dbconnection
    def _send_warnings(self, store):
        """Send warnings to the user who have reached their bounce score
        threshold.

        We send warnings only to the members who have delivery disabled and
        haven't already got a warning in the last
        ``MalingLists.you_are_disabled_warnings_interval`` number of days.
        """
        manager = getUtility(IMembershipManager)
        for member in manager.memberships_pending_warning():
            log.debug('Sending membership disabled warning no. %s to %s due to'
                      ' excessive bounces on %s mailing list',
                      member.total_warnings_sent + 1,
                      member.address,
                      member.mailing_list.display_name)

            send_user_disable_warning(
                member.mailing_list, member.address.email,
                member.preferred_language)
            member.total_warnings_sent += 1
            member.last_warning_sent = now()



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The administrivia rule."""

from email.iterators import typed_subpart_iterator
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.rules import IRule
from public import public
from zope.interface import implementer


# The list of email commands we search for in the Subject header and payload.
# We probably should get this information from the actual implemented
# commands.
EMAIL_COMMANDS = {
    # keyword: (minimum #args, maximum #args)
    'confirm':     (1, 1),
    'help':        (0, 0),
    'info':        (0, 0),
    'lists':       (0, 0),
    'options':     (0, 0),
    'password':    (2, 2),
    'remove':      (0, 0),
    'set':         (3, 3),
    'subscribe':   (0, 3),
    'unsubscribe': (0, 1),
    'who':         (0, 2),
    }


@public
@implementer(IRule)
class Administrivia:
    """The administrivia rule."""

    name = 'administrivia'
    description = _('Catch mis-addressed email commands.')
    record = True

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        # The list must have the administrivia check enabled.
        if not mlist.administrivia:
            return False
        # First check the Subject text.
        lines_to_check = []
        subject = str(msg.get('subject', ''))
        if subject != '':
            lines_to_check.append(subject)
        # Search only the first text/plain subpart of the message.  There's
        # really no good way to find email commands in any other content type.
        for part in typed_subpart_iterator(msg, 'text', 'plain'):
            payload = part.get_payload()
            lines = payload.splitlines()
            # Count lines without using enumerate() because blank lines in the
            # payload don't count against the maximum examined.
            lineno = 0
            for line in lines:
                line = line.strip()
                if len(line) == 0:
                    continue
                lineno += 1
                if lineno > int(config.mailman.email_commands_max_lines):
                    break
                lines_to_check.append(line)
            # Only look at the first text/plain part.
            break
        # For each line we're checking, split the line into words.  Then see
        # if it looks like a command with the min-to-max number of arguments.
        for line in lines_to_check:
            words = [word.lower() for word in line.split()]
            if words[0] not in EMAIL_COMMANDS:
                # This is not an administrivia command.
                continue
            minargs, maxargs = EMAIL_COMMANDS[words[0]]
            if minargs <= len(words) - 1 <= maxargs:
                msgdata['moderation_sender'] = msg.sender
                with _.defer_translation():
                    # This will be translated at the point of use.
                    msgdata.setdefault('moderation_reasons', []).append(
                        _('Message contains administrivia'))
                return True
        return False



# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Add the message to the archives."""

from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.archiver import ArchivePolicy
from mailman.interfaces.handler import IHandler
from public import public
from zope.interface import implementer


@public
@implementer(IHandler)
class ToArchive:
    """Add the message to the archives."""

    name = 'to-archive'
    description = _('Add the message to the archives.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        # Short circuits.
        if (msgdata.get('isdigest') or
                mlist.archive_policy is ArchivePolicy.never):
            return
        # Common practice seems to favor "X-No-Archive: yes".  No other value
        # for this header seems to make sense, so we'll just test for it's
        # presence.  I'm keeping "X-Archive: no" for backwards compatibility.
        if 'x-no-archive' in msg or msg.get('x-archive', '').lower() == 'no':
            return
        # Send the message to the archiver queue.
        config.switchboards['archive'].enqueue(msg, msgdata)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Model for message stores."""

import os
import errno
import pickle

from contextlib import suppress
from mailman.config import config
from mailman.database.transaction import dbconnection
from mailman.interfaces.messages import IMessageStore
from mailman.model.message import Message
from mailman.utilities.email import add_message_hash
from mailman.utilities.filesystem import makedirs, safe_remove
from public import public
from zope.interface import implementer


# It could be very bad if you have already stored files and you change this
# value.  We'd need a script to reshuffle and resplit.
MAX_SPLITS = 2
EMPTYSTRING = ''


@public
@implementer(IMessageStore)
class MessageStore:
    """See `IMessageStore`."""

    @dbconnection
    def add(self, store, message):
        # Ensure that the message has the requisite headers.
        message_ids = message.get_all('message-id', [])
        if len(message_ids) != 1:
            raise ValueError('Exactly one Message-ID header required')
        # Calculate and insert the Message-ID-Hash.
        message_id = message_ids[0]
        if isinstance(message_id, bytes):
            message_id = message_id.decode('ascii')
        # If the Message-ID already exists in the store, don't store it again.
        existing = store.query(Message).filter(
            Message.message_id == message_id).first()
        if existing is not None:
            return None
        hash32 = add_message_hash(message)
        # Calculate the path on disk where we're going to store this message
        # object, in pickled format.
        parts = []
        split = list(hash32)
        while split and len(parts) < MAX_SPLITS:
            parts.append(split.pop(0) + split.pop(0))
        parts.append(hash32)
        relpath = os.path.join(*parts)
        # Store the message in the database.  This relies on the database
        # providing a unique serial number, but to get this information, we
        # have to use a straight insert instead of relying on Elixir to create
        # the object.
        Message(message_id=message_id,
                message_id_hash=hash32,
                path=relpath)
        # Now calculate the full file system path.
        path = os.path.join(config.MESSAGES_DIR, relpath)
        # Write the file to the path, but catch the appropriate exception in
        # case the parent directories don't yet exist.  In that case, create
        # them and try again.
        while True:
            try:
                with open(path, 'wb') as fp:
                    # -1 says to use the highest protocol available.
                    pickle.dump(message, fp, -1)
                    break
            except IOError as error:
                if error.errno != errno.ENOENT:
                    raise
            makedirs(os.path.dirname(path))
        return hash32

    @dbconnection
    def _get_message(self, store, row):
        path = os.path.join(config.MESSAGES_DIR, row.path)
        # The message file may have been externally removed.
        with suppress(FileNotFoundError):
            with open(path, 'rb') as fp:
                return pickle.load(fp)
        # The message is gone.  Delete the entry and return None.
        store.delete(row)
        return None

    @dbconnection
    def get_message_by_id(self, store, message_id):
        row = store.query(Message).filter_by(message_id=message_id).first()
        if row is None:
            return None
        return self._get_message(row)

    @dbconnection
    def get_message_by_hash(self, store, message_id_hash):
        row = store.query(Message).filter_by(
            message_id_hash=message_id_hash).first()
        if row is None:
            return None
        return self._get_message(row)

    @property
    @dbconnection
    def messages(self, store):
        for row in store.query(Message).all():
            yield self._get_message(row)

    @dbconnection
    def delete_message(self, store, message_id):
        row = store.query(Message).filter_by(message_id=message_id).first()
        if row is not None:
            path = os.path.join(config.MESSAGES_DIR, row.path)
            # It's possible that a race condition caused the file system path
            # to already be deleted.
            safe_remove(path)
            store.delete(row)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST API for a mailing list's header matches."""

from mailman.interfaces.action import Action
from mailman.interfaces.mailinglist import IHeaderMatchList
from mailman.rest.helpers import (
    bad_request,
    child,
    CollectionMixin,
    created,
    etag,
    GetterSetter,
    no_content,
    not_found,
    okay,
)
from mailman.rest.validator import enum_validator, regexp_validator, Validator
from public import public


def lowercase(value):
    return str(value).lower()


class _HeaderMatchBase:
    """Common base class."""

    def __init__(self, mlist):
        self._mlist = mlist
        self.header_matches = IHeaderMatchList(self._mlist)

    def _location(self, position):
        return self.api.path_to('lists/{}/header-matches/{}'.format(
            self._mlist.list_id, position))

    def _resource_as_dict(self, header_match):
        """See `CollectionMixin`."""
        resource = dict(
            position=header_match.position,
            header=header_match.header,
            pattern=header_match.pattern,
            self_link=self._location(header_match.position),
            )
        if header_match.chain is not None:
            resource['action'] = header_match.chain
        if header_match.tag is not None:
            resource['tag'] = header_match.tag
        return resource


@public
class HeaderMatch(_HeaderMatchBase):
    """A header match."""

    def __init__(self, mlist, position):
        super().__init__(mlist)
        self._position = position

    def on_get(self, request, response):
        """Get a header match."""
        try:
            header_match = self.header_matches[self._position]
        except IndexError:
            not_found(response, 'No header match at this position: {}'.format(
                      self._position))
        else:
            okay(response, etag(self._resource_as_dict(header_match)))

    def on_delete(self, request, response):
        """Remove a header match."""
        try:
            del self.header_matches[self._position]
        except IndexError:
            not_found(response, 'No header match at this position: {}'.format(
                      self._position))
        else:
            no_content(response)

    def _patch_put(self, request, response, is_optional):
        """Update the header match."""
        try:
            header_match = self.header_matches[self._position]
        except IndexError:
            not_found(response, 'No header match at this position: {}'.format(
                      self._position))
            return
        kws = dict(
            header=lowercase,
            pattern=GetterSetter(regexp_validator),
            position=int,
            action=enum_validator(Action, allow_blank=True),
            tag=lowercase,
            )
        if is_optional:
            # For a PATCH, all attributes are optional.
            kws['_optional'] = kws.keys()
        else:
            # For a PUT, position can remain unchanged; tag and action can be
            # None.
            kws['_optional'] = ('action', 'position', 'tag')
        validator = Validator(**kws)
        try:
            arguments = validator(request)
            missing = object()
            action = arguments.pop('action', missing)
            if action is not missing and action is not None:
                arguments['chain'] = action.name
            elif action is not missing and action is None:
                arguments['chain'] = action
            for key, value in arguments.items():
                setattr(header_match, key, value)
        except ValueError as error:
            bad_request(response, str(error))
            return
        else:
            no_content(response)

    def on_put(self, request, response):
        """Full update of the header match."""
        self._patch_put(request, response, is_optional=False)

    def on_patch(self, request, response):
        """Partial update of the header match."""
        self._patch_put(request, response, is_optional=True)


@public
class HeaderMatches(_HeaderMatchBase, CollectionMixin):
    """The list of all header matches."""

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return list(self.header_matches)

    def on_get(self, request, response):
        """/header-matches"""
        resource = self._make_collection(request)
        okay(response, etag(resource))

    def on_post(self, request, response):
        """Add a header match."""
        validator = Validator(
            header=str,
            pattern=GetterSetter(regexp_validator),
            action=enum_validator(Action, allow_blank=True),
            tag=str,
            _optional=('action', 'tag')
            )
        try:
            arguments = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        action = arguments.pop('action', None)
        if action is not None:
            arguments['chain'] = action.name
        try:
            self.header_matches.append(**arguments)
        except ValueError:
            bad_request(response, b'This header match already exists')
        else:
            header_match = self.header_matches[-1]
            created(response, self._location(header_match.position))

    def on_delete(self, request, response):
        """Delete all header matches for this mailing list."""
        self.header_matches.clear()
        no_content(response)

    @child(r'^(?P<position>\d+)')
    def header_match(self, context, segments, **kw):
        return HeaderMatch(self._mlist, int(kw['position']))

    @child('find')
    def find_matches(self, context, segments, **kw):
        return FindHeaderMatches(self._mlist, **kw)


@public
class FindHeaderMatches(_HeaderMatchBase, CollectionMixin):

    def __init__(self, mlist, **kw):
        self._mlist = mlist

    def on_get(self, request, response):
        return self._find(request, response)

    def on_post(self, request, response):
        return self._find(request, response)

    def _find(self, request, response):
        validator = Validator(
            header=str,
            action=enum_validator(Action),
            tag=str,
            _optional=('action', 'tag', 'header')
           )
        try:
            data = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return

        # Remove any optional pagination elements.
        action = data.pop('action', None)
        if action is not None:
            data['chain'] = action.name
        service = IHeaderMatchList(self._mlist)
        self.header_matches = list(service.filter(**data))

        # Return 404 if no values were found.
        if len(self.header_matches) == 0:
            return not_found(
                response,
                'Cound not find any HeaderMatch for provided search options.')

        resource = self._make_collection(request)
        okay(response, etag(resource))

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return self.header_matches



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The emergency hold rule."""

from mailman.core.i18n import _
from mailman.interfaces.rules import IRule
from public import public
from zope.interface import implementer


@public
@implementer(IRule)
class Emergency:
    """The emergency hold rule."""

    name = 'emergency'

    description = _(
        """The mailing list is in emergency hold and this message was not
        pre-approved by the list administrator.
        """)

    record = True

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        if mlist.emergency and not msgdata.get('moderator_approved'):
            msgdata['moderation_sender'] = msg.sender
            with _.defer_translation():
                # This will be translated at the point of use.
                msgdata.setdefault('moderation_reasons', []).append(
                    _('Emergency moderation is in effect for this list'))
            return True
        return False



"""add_usenet_watermark

Revision ID: 83339e4039da
Revises: 2d2d0ef0828f
Create Date: 2019-07-17 20:41:39.601334

"""

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import exists_in_db


# revision identifiers, used by Alembic.
revision = '83339e4039da'
down_revision = '2d2d0ef0828f'


def upgrade():
    if not exists_in_db(op.get_bind(),
                        'mailinglist',
                        'usenet_watermark'
                        ):
        # SQLite may not have removed it when downgrading.
        op.add_column('mailinglist', sa.Column(
            'usenet_watermark',
            sa.Integer,
            nullable=True))


def downgrade():
    with op.batch_alter_table('mailinglist') as batch_op:
        batch_op.drop_column('usenet_watermark')



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface for describing pipelines."""

from mailman.core.i18n import _, format_reasons
from public import public
from zope.interface import Attribute, Interface


NL = '\n'


# These are thrown but they aren't exceptions so don't inherit from
# mailman.interfaces.errors.MailmanError.  Python requires that they inherit
# from BaseException.
@public
class DiscardMessage(BaseException):
    """The message can be discarded with no further action"""

    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return self.message


@public
class RejectMessage(BaseException):
    """The message will be bounced back to the sender"""

    def __init__(self, message=None, reasons=None, substitutions=None):
        self.message = message
        self.reasons = reasons
        self.substitutions = ({} if substitutions is None else substitutions)

    def __str__(self):
        if self.message is None:
            return _('[No details are available]')
        reasons = (_('[No reasons given]')
                   if self.reasons is None
                   else NL.join(format_reasons(self.reasons)))
        substitutions = self.substitutions.copy()
        substitutions['reasons'] = reasons
        return _(self.message, substitutions)


@public
class IPipeline(Interface):
    """A pipeline of handlers."""

    name = Attribute('Pipeline name; must be unique.')
    description = Attribute('A brief description of this pipeline.')

    def __iter__():
        """Iterate over all the handlers in this pipeline."""



# Sphinx documentation configuration file.
def setup(app):
    # hack the 'exclude_trees' configuration value so that our bounce examples
    # aren't interpreted as documentation.
    app.config.config_values['exclude_trees'][0].extend([
        'tests/bounces',
        'templates',
        ])



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The runner process."""

import os
import sys
import click
import signal
import logging
import traceback

from mailman.config import config
from mailman.core.i18n import _
from mailman.core.initialize import initialize
from mailman.utilities.modules import find_name
from mailman.utilities.options import I18nCommand, validate_runner_spec
from mailman.version import MAILMAN_VERSION_FULL
from public import public


log = None


# Enable coverage if run under the appropriate test suite.
if os.environ.get('COVERAGE_PROCESS_START') is not None:
    import coverage
    coverage.process_startup()


def make_runner(name, slice, range, once=False):
    # The runner name must be defined in the configuration.  Only runner short
    # names are supported.
    runner_config = getattr(config, 'runner.' + name, None)
    if runner_config is None:
        print(_('Undefined runner name: ${name}'), file=sys.stderr)
        # Exit with SIGTERM exit code so the master won't try to restart us.
        sys.exit(signal.SIGTERM)
    class_path = runner_config['class']
    try:
        runner_class = find_name(class_path)
    except ImportError:
        if os.environ.get('MAILMAN_UNDER_MASTER_CONTROL') is not None:
            print(_('Cannot import runner module: ${class_path}'),
                  file=sys.stderr)
            traceback.print_exc()
            sys.exit(signal.SIGTERM)
        else:
            raise
    if once:
        # Subclass to hack in the setting of the stop flag in _do_periodic()
        class Once(runner_class):
            def _do_periodic(self):
                self.stop()
        return Once(name, slice)
    return runner_class(name, slice)


@click.command(
    cls=I18nCommand,
    context_settings=dict(help_option_names=['-h', '--help']),
    help=_("""\
    Start a runner.

    The runner named on the command line is started, and it can either run
    through its main loop once (for those runners that support this) or
    continuously.  The latter is how the master runner starts all its
    subprocesses.

    -r is required unless -l or -h is given, and its argument must be one of
    the names displayed by the -l switch.

    Normally, this script should be started from `mailman start`.  Running it
    separately or with -o is generally useful only for debugging.  When run
    this way, the environment variable $MAILMAN_UNDER_MASTER_CONTROL will be
    set which subtly changes some error handling behavior.
    """))
@click.option(
    '-C', '--config', 'config_file',
    envvar='MAILMAN_CONFIG_FILE',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=_("""\
    Configuration file to use.  If not given, the environment variable
    MAILMAN_CONFIG_FILE is consulted and used if set.  If neither are given, a
    default configuration file is loaded."""))
@click.option(
    '-l', '--list', 'list_runners',
    is_flag=True, is_eager=True, default=False,
    help=_('List the available runner names and exit.'))
@click.option(
    '-o', '--once',
    is_flag=True, default=False,
    help=_("""\
    Run the named runner exactly once through its main loop.  Otherwise, the
    runner runs indefinitely until the process receives a signal.  This is not
    compatible with runners that cannot be run once."""))
@click.option(
    '-r', '--runner', 'runner_spec',
    metavar='runner[:slice:range]',
    callback=validate_runner_spec, default=None,
    help=_("""\

    Start the named runner, which must be one of the strings returned by the -l
    option.

    For runners that manage a queue directory, optional `slice:range` if given
    is used to assign multiple runner processes to that queue.  range is the
    total number of runners for the queue while slice is the number of this
    runner from [0..range).  For runners that do not manage a queue, slice and
    range are ignored.

    When using the `slice:range` form, you must ensure that each runner for the
    queue is given the same range value.  If `slice:runner` is not given, then
    1:1 is used.
    """))
@click.option(
    '-v', '--verbose',
    is_flag=True, default=False,
    help=_('Display more debugging information to the log file.'))
@click.version_option(MAILMAN_VERSION_FULL)
@click.pass_context
@public
def main(ctx, config_file, verbose, list_runners, once, runner_spec):
    # XXX https://github.com/pallets/click/issues/303
    """Start a runner.

    The runner named on the command line is started, and it can either run
    through its main loop once (for those runners that support this) or
    continuously.  The latter is how the master runner starts all its
    subprocesses.

    -r is required unless -l or -h is given, and its argument must be one
    of the names displayed by the -l switch.

    Normally, this script should be started from 'mailman start'.  Running
    it separately or with -o is generally useful only for debugging.  When
    run this way, the environment variable $MAILMAN_UNDER_MASTER_CONTROL
    will be set which subtly changes some error handling behavior.
    """
    global log

    if runner_spec is None and not list_runners:
        ctx.fail(_('No runner name given.'))

    # Initialize the system.  Honor the -C flag if given.
    initialize(config_file, verbose)
    log = logging.getLogger('mailman.runner')
    if verbose:
        console = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter(config.logging.root.format,
                                      config.logging.root.datefmt)
        console.setFormatter(formatter)
        logging.getLogger().addHandler(console)
        logging.getLogger().setLevel(logging.DEBUG)

    if list_runners:
        descriptions = {}
        for section in config.runner_configs:
            ignore, dot, shortname = section.name.rpartition('.')
            ignore, dot, classname = getattr(section, 'class').rpartition('.')
            descriptions[shortname] = classname
        longest = max(len(name) for name in descriptions)
        for shortname in sorted(descriptions):
            classname = descriptions[shortname]
            spaces = longest - len(shortname)
            name = (' ' * spaces) + shortname       # noqa: F841
            print(_('${name} runs ${classname}'))
        sys.exit(0)

    runner = make_runner(*runner_spec, once=once)
    runner.set_signals()
    # Now start up the main loop
    log.info('{} runner started.'.format(runner.name))
    runner.run()
    log.info('{} runner exiting.'.format(runner.name))
    sys.exit(runner.status)



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""A wrapper around passlib."""

from mailman.config.config import load_external
from mailman.interfaces.configuration import ConfigurationUpdatedEvent
from passlib.context import CryptContext
from passlib.exc import UnknownHashError
from public import public


class PasswordContext:
    def __init__(self, config):
        """Create a password context for hashing and verification.

        :param config: The `IConfiguration` instance.
        """
        config_string = load_external(config.passwords.configuration)
        self._context = CryptContext.from_string(config_string)

    def encrypt(self, secret):
        """Return the secret, hashed using the current password context.

        :param secret: The plain text password.
        :type secret: string
        :return: The hashed secret.
        :rtype: string
        """
        return self._context.encrypt(secret)

    def verify(self, password, hashed):
        """Verify the hashed password and return the updated hash.

        This is essentially a wrapper around
        `passlib.CryptContext.verify_and_update()` using only the first two
        arguments.

        :param password: The plain text secret provided by the user.
        :type password:
        :param hashed: The hash string to compare to.
        :type hashed: string
        :return: 2-tuple where the first element is a flag indicating whether
            the password verified or not, and the second value whether the
            existing hash needs to be replaced (a str if so, else None).
        :rtype: 2-tuple
        """
        try:
            return self._context.verify_and_update(password, hashed)
        except UnknownHashError:
            return (False, password)


@public
def handle_ConfigurationUpdatedEvent(event):
    if isinstance(event, ConfigurationUpdatedEvent):
        # Just reset the password context.
        event.config.password_context = PasswordContext(event.config)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST for members."""

import logging

from lazr.config import as_boolean
from mailman.app.membership import add_member
from mailman.interfaces.action import Action
from mailman.interfaces.address import IAddress, InvalidEmailAddressError
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import (
    AlreadySubscribedError,
    DeliveryMode,
    DeliveryStatus,
    MemberRole,
    MembershipError,
    MembershipIsBannedError,
    MissingPreferredAddressError,
)
from mailman.interfaces.subscriptions import (
    ISubscriptionManager,
    ISubscriptionService,
    RequestRecord,
    SubscriptionPendingError,
    TokenOwner,
)
from mailman.interfaces.user import IUser, UnverifiedAddressError
from mailman.interfaces.usermanager import IUserManager
from mailman.rest.helpers import (
    accepted,
    bad_request,
    child,
    CollectionMixin,
    conflict,
    created,
    etag,
    no_content,
    not_found,
    NotFound,
    okay,
)
from mailman.rest.preferences import Preferences, ReadOnlyPreferences
from mailman.rest.validator import (
    enum_validator,
    list_of_strings_validator,
    subscriber_validator,
    Validator,
)
from operator import attrgetter
from public import public
from uuid import UUID
from zope.component import getUtility


log = logging.getLogger('mailman.rest')


class _MemberBase(CollectionMixin):
    """Shared base class for member representations."""

    @property
    def all_fields(self):
        """Get a mapping of all the supported fields for a Member resource.

        Keys are the name of the fields and values are callables that return
        values of those fields.
        """
        # The member will always have a member id and an address id.  It will
        # only have a user id if the address is linked to a user.
        # E.g. nonmembers we've only seen via postings to lists they are not
        # subscribed to will not have a user id.  The user_id and the
        # member_id are UUIDs.  In API 3.0 we use the integer equivalent of
        # the UID in the URL, but in API 3.1 we use the hex equivalent.  See
        # issue #121 for details.
        return {
            'address': self._get_address,
            'bounce_score': attrgetter('bounce_score'),
            'last_bounce_received': attrgetter('last_bounce_received'),
            'last_warning_sent': attrgetter('last_warning_sent'),
            'total_warnings_sent': attrgetter('total_warnings_sent'),
            'delivery_mode': attrgetter('delivery_mode'),
            'email': attrgetter('address.email'),
            'list_id': attrgetter('list_id'),
            'subscription_mode': attrgetter('subscription_mode'),
            'role': attrgetter('role.name'),
            'user': self._get_user,
            'moderation_action': attrgetter('moderation_action'),
            'display_name': attrgetter('display_name'),
            'self_link': self._get_self_link,
            'member_id': self._get_member_id,
            }

    def _get_member_id(self, member):
        """Get member_id."""
        return self.api.from_uuid(member.member_id)

    def _get_address(self, member):
        """Get url to member's addresses."""
        return self.api.path_to(
                'addresses/{}'.format(member.address.email))

    def _get_user(self, member):
        """Get url to member's user if one exists."""
        user = member.user
        if user:
            user_id = self.api.from_uuid(user.user_id)
            return self.api.path_to('users/{}'.format(user_id))
        return None

    def _get_self_link(self, member):
        """Get self_link to member resource."""
        member_id = self.api.from_uuid(member.member_id)
        return self.api.path_to('members/{}'.format(member_id))

    def _resource_as_dict(self, member, fields=None):
        """See `CollectionMixin`."""
        if fields is None:
            fields = self.all_fields.keys()

        response = {}
        for field in fields:
            value_getter = self.all_fields.get(field, None)
            if value_getter is None:
                raise ValueError(
                    'Unknown field "{}" for Member resource.'
                    ' Allowed fields are: {}'.format(
                        field, ', '.join(self.all_fields.keys())))
            value = value_getter(member)
            if value is not None:
                response[field] = value
        return response

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return list(getUtility(ISubscriptionService))

    def on_get(self, request, response):
        """/members"""
        validator = Validator(
            fields=list_of_strings_validator,
            count=int,
            page=int,
            _optional=['fields', 'count', 'page'],
            )
        try:
            data = validator(request)
        except ValueError as ex:
            bad_request(response, str(ex))
            return
        fields = data.get('fields', None)

        try:
            resource = self._make_collection(request, fields)
        except ValueError as ex:
            bad_request(response, str(ex))
            return
        okay(response, etag(resource))


@public
class MemberCollection(_MemberBase):
    """Abstract class for supporting submemberships.

    This is used for example to return a resource representing all the
    memberships of a mailing list, or all memberships for a specific email
    address.
    """
    def _get_collection(self, request):
        """See `CollectionMixin`."""
        raise NotImplementedError


@public
class AMember(_MemberBase):
    """A member."""

    def __init__(self, member_id):
        # The member_id is either the member's UUID or the string
        # representation of the member's UUID.
        service = getUtility(ISubscriptionService)
        self._member_id = member_id
        self._member = (None if member_id is None
                        else service.get_member(member_id))

    def on_get(self, request, response):
        """Return a single member end-point."""
        if self._member is None:
            not_found(response)
        else:
            okay(response, self._resource_as_json(self._member))

    @child()
    def preferences(self, context, segments):
        """/members/<id>/preferences"""
        if len(segments) != 0:
            return NotFound(), []
        if self._member is None:
            return NotFound(), []
        member_id = self.api.from_uuid(self._member_id)
        child = Preferences(
            self._member.preferences, 'members/{}'.format(member_id))
        return child, []

    @child()
    def all(self, context, segments):
        """/members/<id>/all/preferences"""
        if len(segments) == 0:
            return NotFound(), []
        if self._member is None:
            return NotFound(), []
        member_id = self.api.from_uuid(self._member_id)
        child = ReadOnlyPreferences(
            self._member, 'members/{}/all'.format(member_id))
        return child, []

    def on_delete(self, request, response):
        """Delete the member (i.e. unsubscribe)."""
        # Leaving a list is a bit different than deleting a moderator or
        # owner.  Handle the former case first.  For now too, we will not send
        # an admin or user notification.
        if self._member is None:
            not_found(response)
            return
        mlist = getUtility(IListManager).get_by_list_id(self._member.list_id)
        if self._member.role is MemberRole.member:
            try:
                values = Validator(
                    pre_confirmed=as_boolean,
                    pre_approved=as_boolean,
                    _optional=('pre_confirmed', 'pre_approved'),
                    )(request)
            except ValueError as error:
                bad_request(response, str(error))
                return
            manager = ISubscriptionManager(mlist)
            # XXX(maxking): For backwards compat, we are going to keep
            # pre-confirmed to be "True" by defualt instead of "False", that it
            # should be. Any, un-authenticated requests should manually specify
            # that it is *not* confirmed by the user.
            if 'pre_confirmed' in values:
                pre_confirmed = values.get('pre_confirmed')
            else:
                pre_confirmed = True
            token, token_owner, member = manager.unregister(
                self._member.address,
                pre_approved=values.get('pre_approved'),
                pre_confirmed=pre_confirmed)
            if member is None:
                assert token is None
                assert token_owner is TokenOwner.no_one
                no_content(response)
            else:
                assert token is not None
                content = dict(token=token, token_owner=token_owner.name)
                accepted(response, etag(content))
        else:
            self._member.unsubscribe()
            no_content(response)

    def on_patch(self, request, response):
        """Patch the membership.

        This is how subscription changes are done.
        """
        if self._member is None:
            not_found(response)
            return
        try:
            values = Validator(
                address=str,
                delivery_mode=enum_validator(DeliveryMode),
                moderation_action=enum_validator(Action, allow_blank=True),
                _optional=('address', 'delivery_mode', 'moderation_action'),
                )(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        if 'address' in values:
            email = values['address']
            address = getUtility(IUserManager).get_address(email)
            if address is None:
                bad_request(response, b'Address not registered')
                return
            try:
                self._member.address = address
            except (MembershipError, UnverifiedAddressError) as error:
                bad_request(response, str(error))
                return
        if 'delivery_mode' in values:
            self._member.preferences.delivery_mode = values['delivery_mode']
        if 'moderation_action' in values:
            self._member.moderation_action = values['moderation_action']
        no_content(response)


@public
class AllMembers(_MemberBase):
    """The members."""

    def on_post(self, request, response):
        """Create a new member."""
        try:
            validator = Validator(
                list_id=str,
                subscriber=subscriber_validator(self.api),
                display_name=str,
                delivery_mode=enum_validator(DeliveryMode),
                delivery_status=enum_validator(DeliveryStatus),
                role=enum_validator(MemberRole),
                pre_verified=as_boolean,
                pre_confirmed=as_boolean,
                pre_approved=as_boolean,
                invitation=as_boolean,
                send_welcome_message=as_boolean,
                _optional=('delivery_mode', 'display_name', 'role',
                           'pre_verified', 'pre_confirmed', 'pre_approved',
                           'invitation', 'send_welcome_message',
                           'delivery_status'))
            arguments = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        # Dig the mailing list out of the arguments.
        list_id = arguments.pop('list_id')
        mlist = getUtility(IListManager).get_by_list_id(list_id)
        if mlist is None:
            bad_request(response, b'No such list')
            return
        # Figure out what kind of subscriber is being registered.  Either it's
        # a user via their preferred email address or it's an explicit address.
        # If it's a UUID, then it must be associated with an existing user.
        subscriber = arguments.pop('subscriber')
        user_manager = getUtility(IUserManager)
        # We use the display name if there is one.
        display_name = arguments.pop('display_name', '')
        if isinstance(subscriber, UUID):
            user = user_manager.get_user_by_id(subscriber)
            if user is None:
                bad_request(response, b'No such user')
                return
            subscriber = user
        else:
            # This must be an email address.  See if there's an existing
            # address object associated with this email.
            address = user_manager.get_address(subscriber)
            if address is None:
                # Create a new address, which of course will not be validated.
                address = user_manager.create_address(
                    subscriber, display_name)
            subscriber = address
        # What role are we subscribing?  Regular members go through the
        # subscription policy workflow while owners, moderators, and
        # nonmembers go through the legacy API for now.
        role = arguments.pop('role', MemberRole.member)
        if role is MemberRole.member:
            # Get the pre_ flags for the subscription workflow.
            pre_verified = arguments.pop('pre_verified', False)
            pre_confirmed = arguments.pop('pre_confirmed', False)
            pre_approved = arguments.pop('pre_approved', False)
            invitation = arguments.pop('invitation', False)
            send_welcome_message = arguments.pop('send_welcome_message', None)
            delivery_status = arguments.pop('delivery_status', None)
            delivery_mode = arguments.pop('delivery_mode', None)
            # Now we can run the registration process until either the
            # subscriber is subscribed, or the workflow is paused for
            # verification, confirmation, or approval.
            registrar = ISubscriptionManager(mlist)
            try:
                token, token_owner, member = registrar.register(
                    subscriber,
                    pre_verified=pre_verified,
                    pre_confirmed=pre_confirmed,
                    pre_approved=pre_approved,
                    invitation=invitation,
                    send_welcome_message=send_welcome_message,
                    delivery_mode=delivery_mode,
                    delivery_status=delivery_status)
            except AlreadySubscribedError:
                conflict(response, b'Member already subscribed')
                return
            except MissingPreferredAddressError:
                bad_request(response, b'User has no preferred address')
                return
            except MembershipIsBannedError:
                bad_request(response, b'Membership is banned')
                return
            except InvalidEmailAddressError:
                bad_request(response, b'List posting address cannot subscribe')
                return
            except SubscriptionPendingError:
                conflict(response, b'Subscription request already pending')
                return
            except Exception as e:
                log.exception(e)
                bad_request(response, str(e))
                return
            if token is None:
                assert token_owner is TokenOwner.no_one, token_owner
                # The subscription completed.  Let's get the resulting member
                # and return the location to the new member.  Member ids are
                # UUIDs and need to be converted to URLs because JSON doesn't
                # directly support UUIDs.
                member_id = self.api.from_uuid(member.member_id)
                location = self.api.path_to('members/{}'.format(member_id))
                created(response, location)
                return
            # The member could not be directly subscribed because there are
            # some out-of-band steps that need to be completed.  E.g. the user
            # must confirm their subscription or the moderator must approve
            # it.  In this case, an HTTP 202 Accepted is exactly the code that
            # we should use, and we'll return both the confirmation token and
            # the "token owner" so the client knows who should confirm it.
            assert token is not None, token
            assert token_owner is not TokenOwner.no_one, token_owner
            assert member is None, member
            content = dict(token=token, token_owner=token_owner.name)
            accepted(response, etag(content))
            return
        # 2015-04-15 BAW: We're subscribing some role other than a regular
        # member.  Use the legacy API for this for now.
        assert role in (MemberRole.owner,
                        MemberRole.moderator,
                        MemberRole.nonmember)
        # 2015-04-15 BAW: We're limited to using an email address with this
        # legacy API, so if the subscriber is a user, the user must have a
        # preferred address, which we'll use, even though it will subscribe
        # the explicit address.  It is an error if the user does not have a
        # preferred address.
        #
        # If the subscriber is an address object, just use that.
        if IUser.providedBy(subscriber):
            if subscriber.preferred_address is None:
                bad_request(response, b'User without preferred address')
                return
            email = subscriber.preferred_address.email
        else:
            assert IAddress.providedBy(subscriber)
            email = subscriber.email
        delivery_mode = arguments.pop('delivery_mode', DeliveryMode.regular)
        delivery_status = arguments.pop('delivery_status', None)
        record = RequestRecord(
            email, display_name, delivery_mode,
            delivery_status=delivery_status)
        try:
            member = add_member(mlist, record, role)
        except MembershipIsBannedError:
            bad_request(response, b'Membership is banned')
            return
        except AlreadySubscribedError:
            bad_request(response,
                        '{} is already an {} of {}'.format(
                            email, role.name, mlist.fqdn_listname))
            return
        except InvalidEmailAddressError:
            bad_request(response, b'List posting address cannot be added')
            return
        # The subscription completed.  Let's get the resulting member
        # and return the location to the new member.  Member ids are
        # UUIDs and need to be converted to URLs because JSON doesn't
        # directly support UUIDs.
        member_id = self.api.from_uuid(member.member_id)
        location = self.api.path_to('members/{}'.format(member_id))
        created(response, location)


class _FoundMembers(MemberCollection):
    """The found members collection."""

    def __init__(self, members, api):
        super().__init__()
        self._members = members
        self.api = api

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return self._members


@public
class FindMembers(_MemberBase):
    """/members/find"""

    def on_get(self, request, response):
        return self._find(request, response)

    def on_post(self, request, response):
        return self._find(request, response)

    def _find(self, request, response):
        """Find a member"""
        service = getUtility(ISubscriptionService)
        validator = Validator(
            list_id=str,
            subscriber=str,
            role=enum_validator(MemberRole),
            moderation_action=enum_validator(Action, allow_blank=True),
            # preferences.
            delivery_mode=enum_validator(DeliveryMode),
            delivery_status=enum_validator(DeliveryStatus),
            # Allow pagination.
            page=int,
            count=int,
            fields=list_of_strings_validator,
            _optional=(
                'list_id', 'subscriber', 'role', 'moderation_action',
                'delivery_mode', 'delivery_status',
                'page', 'count', 'fields'))
        try:
            data = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
        else:
            # Remove any optional pagination query elements; they will be
            # handled later.
            data.pop('page', None)
            data.pop('count', None)
            fields = data.pop('fields', None)
            members = service.find_members(**data)
            resource = _FoundMembers(members, self.api)
            try:
                collection = resource._make_collection(request, fields)
            except ValueError as ex:
                bad_request(response, str(ex))
                return
            okay(response, etag(collection))



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""URI templates."""

from mailman.interfaces.template import ALL_TEMPLATES, ITemplateManager
from mailman.rest.helpers import (
    bad_request,
    CollectionMixin,
    etag,
    no_content,
    not_found,
    okay,
)
from mailman.rest.validator import Validator
from operator import attrgetter
from public import public
from zope.component import getUtility


class _URIBase(CollectionMixin):
    def __init__(self, context):
        self._context = context

    def _resource_as_dict(self, template):
        resource = dict(
            uri=template.uri,
            name=template.name,
            self_link=self.api.path_to('{}/uris/{}'.format(
                self._prefix, template.name)),
            )
        if template.username is not None and template.password is not None:
            resource['username'] = template.username
            resource['password'] = template.password
        return resource

    def _get_collection(self, request):
        manager = getUtility(ITemplateManager)
        collection = []
        for uri in self.URIs:
            template = manager.raw(uri, self._raw_context)
            if template is not None:
                collection.append(template)
        return sorted(collection, key=attrgetter('name'))

    def on_get(self, request, response):
        resource = self._make_collection(request)
        resource['self_link'] = self.api.path_to(
            '{}/uris'.format(self._prefix))
        okay(response, etag(resource))

    def _patch_put(self, request, response, is_optional):
        kws = {uri: str for uri in self.URIs}
        optionals = ['username', 'password']
        if is_optional:
            optionals.extend(self.URIs)
        # When PATCHing or PUTing all uris, a single optional
        # username/password applies to them all.
        kws['username'] = str
        kws['password'] = str
        kws['_optional'] = optionals
        try:
            arguments = Validator(**kws)(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        username = arguments.pop('username', None)
        password = arguments.pop('password', None)
        if not username and not password:
            # Normalize arguments.
            set_kws = {}
        elif username and password:
            # It's fine if both are specified.
            set_kws = dict(username=username, password=password)
        else:
            bad_request(response,
                        'Specify both username and password, or neither')
            return
        manager = getUtility(ITemplateManager)
        for key, value in arguments.items():
            if len(value) == 0:
                # The empty string is equivalent to DELETE.  Yeah, this isn't
                # very RESTful, but practicality beats purity.
                manager.delete(key, self._raw_context)
            else:
                manager.set(key, self._raw_context, value, **set_kws)
        no_content(response)

    def on_put(self, request, response):
        self._patch_put(request, response, is_optional=False)

    def on_patch(self, request, response):
        self._patch_put(request, response, is_optional=True)

    def on_delete(self, request, response):
        manager = getUtility(ITemplateManager)
        for uri in self.URIs:
            manager.delete(uri, self._raw_context)
        no_content(response)


class _ListURIBase(_URIBase):
    def __init__(self, context):
        super().__init__(context)
        self._raw_context = context.list_id
        self._prefix = 'lists/{}'.format(context.list_id)


@public
class AllListURIs(_ListURIBase):
    URIs = [name for name in ALL_TEMPLATES if name.startswith('list:')]

    def __init__(self, context):
        super().__init__(context)


@public
class AListURI(_ListURIBase):
    def __init__(self, context, template):
        super().__init__(context)
        self.URIs = [template]
        self._template = template

    def on_get(self, request, response):
        template = getUtility(ITemplateManager).raw(
            self._template, self._raw_context)
        if template is None:
            not_found(response)
        else:
            resource = dict(uri=template.uri)
            resource['self_link'] = self.api.path_to(
                '{}/uris/{}'.format(self._prefix, self._template))
            okay(response, etag(resource))


class _DomainURIBase(_URIBase):
    def __init__(self, context):
        super().__init__(context)
        self._raw_context = context.mail_host
        self._prefix = 'domains/{}'.format(context.mail_host)


@public
class AllDomainURIs(_DomainURIBase):
    URIs = [name for name in ALL_TEMPLATES
            if name.startswith('list:') or name.startswith('domain:')]


@public
class ADomainURI(_DomainURIBase):
    def __init__(self, context, template):
        super().__init__(context)
        self.URIs = [template]
        self._template = template

    def on_get(self, request, response):
        template = getUtility(ITemplateManager).raw(
            self._template, self._raw_context)
        if template is None:
            not_found(response)
        else:
            resource = dict(uri=template.uri)
            resource['self_link'] = self.api.path_to(
                '{}/uris/{}'.format(self._prefix, self._template))
            okay(response, etag(resource))


class _SiteURIBase(_URIBase):
    def __init__(self):
        super().__init__(None)
        self._raw_context = None
        self._prefix = ''


@public
class AllSiteURIs(_SiteURIBase):
    URIs = [name for name in ALL_TEMPLATES]


@public
class ASiteURI(_SiteURIBase):
    def __init__(self, template):
        super().__init__()
        self.URIs = [template]
        self._template = template

    def on_get(self, request, response):
        template = getUtility(ITemplateManager).raw(self._template, None)
        if template is None:
            not_found(response)
        else:
            resource = dict(uri=template.uri)
            resource['self_link'] = self.api.path_to(
                'uris/{}'.format(self._template))
            okay(response, etag(resource))



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Switchboard tests."""

import os
import unittest

from mailman.config import config
from mailman.testing.helpers import (
    LogFileMark,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from unittest.mock import patch


class TestSwitchboard(unittest.TestCase):
    layer = ConfigLayer

    def test_log_exception_in_finish(self):
        # If something bad happens in .finish(), the traceback should get
        # logged.  LP: #1165589.
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Message-ID: <ant>

""")
        switchboard = config.switchboards['shunt']
        # Enqueue the message.
        filebase = switchboard.enqueue(msg)
        error_log = LogFileMark('mailman.error')
        msg, data = switchboard.dequeue(filebase)
        # Now, cause .finish() to throw an exception.
        with patch('mailman.core.switchboard.os.rename',
                   side_effect=OSError('Oops!')):
            switchboard.finish(filebase, preserve=True)
        traceback = error_log.read().splitlines()
        self.assertEqual(traceback[1], 'Traceback (most recent call last):')
        self.assertEqual(traceback[-1], 'OSError: Oops!')

    def test_no_bak_but_pck(self):
        # if there is no .bak file but a .pck with the same filebase,
        # .finish() should handle the .pck.
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Message-ID: <ant>

""")
        switchboard = config.switchboards['shunt']
        # Enqueue the message.
        filebase = switchboard.enqueue(msg)
        # Now call .finish() without first dequeueing.
        switchboard.finish(filebase, preserve=True)
        # And ensure the file got preserved.
        bad_dir = config.switchboards['bad'].queue_directory
        psvfile = os.path.join(bad_dir, filebase + '.psv')
        self.assertTrue(os.path.isfile(psvfile))



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test domains."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.database.transaction import dbconnection, transaction
from mailman.interfaces.domain import (
    DomainCreatedEvent,
    DomainCreatingEvent,
    DomainDeletedEvent,
    DomainDeletingEvent,
    IDomainManager,
)
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.usermanager import IUserManager
from mailman.model.domain import Domain
from mailman.testing.helpers import event_subscribers
from mailman.testing.layers import ConfigLayer
from sqlalchemy.exc import IntegrityError
from zope.component import getUtility


class TestDomainManager(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._events = []
        self._manager = getUtility(IDomainManager)

    def _record_event(self, event):
        self._events.append(event)

    def test_create_domain_event(self):
        # Test that creating a domain in the domain manager propagates the
        # expected events.
        with event_subscribers(self._record_event):
            domain = self._manager.add('example.org')
        self.assertEqual(len(self._events), 2)
        self.assertIsInstance(self._events[0], DomainCreatingEvent)
        self.assertEqual(self._events[0].mail_host, 'example.org')
        self.assertIsInstance(self._events[1], DomainCreatedEvent)
        self.assertEqual(self._events[1].domain, domain)

    def test_delete_domain_event(self):
        # Test that deleting a domain in the domain manager propagates the
        # expected event.
        domain = self._manager.add('example.org')
        with event_subscribers(self._record_event):
            self._manager.remove('example.org')
        self.assertEqual(len(self._events), 2)
        self.assertIsInstance(self._events[0], DomainDeletingEvent)
        self.assertEqual(self._events[0].domain, domain)
        self.assertIsInstance(self._events[1], DomainDeletedEvent)
        self.assertEqual(self._events[1].mail_host, 'example.org')

    def test_lookup_missing_domain(self):
        # Like dictionaries, getitem syntax raises KeyError on missing domain.
        with self.assertRaises(KeyError):
            self._manager['doesnotexist.com']

    def test_delete_missing_domain(self):
        # Trying to delete a missing domain gives you a KeyError.
        self.assertRaises(KeyError, self._manager.remove, 'doesnotexist.com')

    def test_domain_creation_no_default_owners(self):
        # If a domain is created without owners, then it has none.
        domain = self._manager.add('example.org')
        self.assertEqual(len(domain.owners), 0)

    def test_domain_creation_no_alias(self):
        # If a domain is created without an alias, then it has none.
        domain = self._manager.add('example.org')
        self.assertIsNone(domain.alias_domain)

    def test_domain_creation_with_alias(self):
        # If a domain is created with an alias then it has one.
        domain = self._manager.add('example.org', alias_domain='x.example.org')
        self.assertEqual(domain.alias_domain, 'x.example.org')

    def test_domain_creation_with_owner(self):
        # You can create a new domain with a single owner.
        domain = self._manager.add('example.org', owners=['anne@example.org'])
        self.assertEqual(len(domain.owners), 1)
        self.assertEqual(domain.owners[0].addresses[0].email,
                         'anne@example.org')

    def test_domain_creation_with_owners(self):
        # You can create a new domain with multiple owners.
        domain = self._manager.add(
            'example.org', owners=['anne@example.org',
                                   'bart@example.net'])
        self.assertEqual(len(domain.owners), 2)
        self.assertEqual(
            sorted(owner.addresses[0].email for owner in domain.owners),
            ['anne@example.org', 'bart@example.net'])

    def test_domain_creation_creates_new_users(self):
        # Domain creation with existing users does not create new users, but
        # any user which doesn't yet exist (and is linked to the given
        # address), gets created.
        user_manager = getUtility(IUserManager)
        user_manager.make_user('anne@example.com')
        user_manager.make_user('bart@example.com')
        domain = self._manager.add(
            'example.org', owners=['anne@example.com',
                                   'bart@example.com',
                                   'cris@example.com'])
        self.assertEqual(len(domain.owners), 3)
        self.assertEqual(
            sorted(owner.addresses[0].email for owner in domain.owners),
            ['anne@example.com', 'bart@example.com', 'cris@example.com'])
        # Now cris exists as a user.
        self.assertIsNotNone(user_manager.get_user('cris@example.com'))

    def test_domain_creation_with_users(self):
        # Domains can be created with IUser objects.
        user_manager = getUtility(IUserManager)
        anne = user_manager.make_user('anne@example.com')
        bart = user_manager.make_user('bart@example.com')
        domain = self._manager.add('example.org', owners=[anne, bart])
        self.assertEqual(len(domain.owners), 2)
        self.assertEqual(
            sorted(owner.addresses[0].email for owner in domain.owners),
            ['anne@example.com', 'bart@example.com'])
        def sort_key(owner):                               # noqa: E306
            return owner.addresses[0].email
        self.assertEqual(sorted(domain.owners, key=sort_key), [anne, bart])

    def test_add_domain_owner(self):
        # Domain owners can be added after the domain is created.
        domain = self._manager.add('example.org')
        self.assertEqual(len(domain.owners), 0)
        domain.add_owner('anne@example.org')
        self.assertEqual(len(domain.owners), 1)
        self.assertEqual(domain.owners[0].addresses[0].email,
                         'anne@example.org')

    def test_add_multiple_domain_owners(self):
        # Multiple domain owners can be added after the domain is created.
        domain = self._manager.add('example.org')
        self.assertEqual(len(domain.owners), 0)
        domain.add_owners(['anne@example.org', 'bart@example.net'])
        self.assertEqual(len(domain.owners), 2)
        self.assertEqual([owner.addresses[0].email for owner in domain.owners],
                         ['anne@example.org', 'bart@example.net'])

    def test_remove_domain_owner(self):
        # Domain onwers can be removed.
        domain = self._manager.add(
            'example.org', owners=['anne@example.org',
                                   'bart@example.net'])
        domain.remove_owner('anne@example.org')
        self.assertEqual(len(domain.owners), 1)
        self.assertEqual([owner.addresses[0].email for owner in domain.owners],
                         ['bart@example.net'])

    def test_remove_missing_owner(self):
        # Users which aren't owners can't be removed.
        domain = self._manager.add(
            'example.org', owners=['anne@example.org',
                                   'bart@example.net'])
        self.assertRaises(ValueError, domain.remove_owner, 'cris@example.org')
        self.assertEqual(len(domain.owners), 2)
        self.assertEqual([owner.addresses[0].email for owner in domain.owners],
                         ['anne@example.org', 'bart@example.net'])


class TestDomain(unittest.TestCase):
    layer = ConfigLayer

    @dbconnection
    def test_unique_mail_host(self, store):
        domain = Domain('abc')
        # Creates the first domain and commit it right away
        with transaction():
            store.add(domain)

        # Now creating another Domain with same mail_host and committing it
        with self.assertRaises(IntegrityError):
            with transaction():
                store.add(Domain('abc'))


class TestDomainLifecycleEvents(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._domainmanager = getUtility(IDomainManager)
        self._org = self._domainmanager.add('example.net')
        self._net = self._domainmanager.add('example.org')

    def test_lists_are_deleted_when_domain_is_deleted(self):
        # When a domain is deleted, all the mailing lists in that domain are
        # also deleted.
        create_list('ant@example.net')
        create_list('bee@example.net')
        cat = create_list('cat@example.org')
        dog = create_list('dog@example.org')
        ewe = create_list('ewe@example.com')
        fly = create_list('fly@example.com')
        listmanager = getUtility(IListManager)
        self._domainmanager.remove('example.net')
        self.assertEqual(listmanager.get('ant@example.net'), None)
        self.assertEqual(listmanager.get('bee@example.net'), None)
        self.assertEqual(listmanager.get('cat@example.org'), cat)
        self.assertEqual(listmanager.get('dog@example.org'), dog)
        self.assertEqual(listmanager.get('ewe@example.com'), ewe)
        self.assertEqual(listmanager.get('fly@example.com'), fly)
        self._domainmanager.remove('example.org')
        self.assertEqual(listmanager.get('cat@example.org'), None)
        self.assertEqual(listmanager.get('dog@example.org'), None)
        self.assertEqual(listmanager.get('ewe@example.com'), ewe)
        self.assertEqual(listmanager.get('fly@example.com'), fly)



# Copyright (C) 2006-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Transactional support."""

from contextlib import contextmanager
from mailman.config import config
from public import public


@public
@contextmanager
def transaction():
    """Context manager for ensuring the transaction is complete."""
    try:
        yield
    except:                                              # noqa: E722
        config.db.abort()
        raise
    else:
        config.db.commit()


@public
def transactional(function):
    """Decorator for transactional support.

    When the function this decorator wraps exits cleanly, the current
    transaction is committed.  When it exits uncleanly (i.e. because of an
    exception, the transaction is aborted.

    Either way, the current transaction is completed.
    """
    def wrapper(*args, **kws):
        try:
            rtn = function(*args, **kws)
            config.db.commit()
            return rtn
        except:                                  # noqa: E722 pragma: nocover
            config.db.abort()
            raise
    return wrapper


@public
def api_transaction(function):
    """Decorator for transactional support in REST API.

    Difference between this and :py:func:`transactional` is just that this
    one also closes the session. This is used in REST API to ensure that
    no state is retained after a request-response cycle.
    """
    def wrapper(*args, **kws):
        try:
            rtn = function(*args, **kws)
            config.db.commit()
            config.db.close_session()
            return rtn
        except:                                  # noqa: E722 pragma: nocover
            config.db.abort()
            config.db.close_session()
            raise
    return wrapper


@public
@contextmanager
def flush():
    """Context manager for flushing SQLAlchemy.

    We need this for SA whereas we didn't need it for Storm because the latter
    did auto-reloads.  However, in SA this is needed when we add or delete
    objects from the database.  Use it when you need the id after adding, or
    when you want to be sure the object won't be found after a delete.

    This is lighter weight than committing the transaction.
    """
    yield
    config.db.store.flush()


@public
def dbconnection(function):
    """Decorator for getting at the database connection.

    Use this to avoid having to access the global `config.db.store`
    attribute.  This calls the function with `store` as the first argument.
    """
    def wrapper(*args, **kws):
        # args[0] is self, if there is one.
        if len(args) > 0:
            return function(args[0], config.db.store, *args[1:], **kws)
        else:
            return function(config.db.store, **kws)
    return wrapper



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The default built-in starting chain."""

import logging

from mailman.chains.base import Link
from mailman.core.i18n import _
from mailman.interfaces.chain import IChain, LinkAction
from public import public
from zope.interface import implementer


log = logging.getLogger('mailman.vette')


@public
@implementer(IChain)
class BuiltInChain:
    """Default built-in chain."""

    name = 'default-posting-chain'
    description = _('The built-in moderation chain.')

    _link_descriptions = (
        # First check DMARC.  For a reject or discard, the rule hits and we
        # jump to the moderation chain to do the action.  Otherwise, the rule
        # misses buts sets msgdata['dmarc'] for the handler.
        ('dmarc-mitigation', LinkAction.jump, 'dmarc'),
        # Discard emails with no valid senders.
        ('no-senders', LinkAction.jump, 'discard'),
        ('approved', LinkAction.jump, 'accept'),
        ('emergency', LinkAction.jump, 'hold'),
        ('loop', LinkAction.jump, 'discard'),
        # Discard emails from banned addresses.
        ('banned-address', LinkAction.jump, 'discard'),
        # Determine whether the member or nonmember has an action shortcut.
        ('member-moderation', LinkAction.jump, 'moderation'),
        # Take a detour through the header matching chain.
        ('truth', LinkAction.detour, 'header-match'),
        # Check for nonmember moderation.
        ('nonmember-moderation', LinkAction.jump, 'moderation'),
        # Do all of the following before deciding whether to hold the message.
        ('administrivia', LinkAction.defer, None),
        ('implicit-dest', LinkAction.defer, None),
        ('max-recipients', LinkAction.defer, None),
        ('max-size', LinkAction.defer, None),
        ('news-moderation', LinkAction.defer, None),
        ('no-subject', LinkAction.defer, None),
        ('digests', LinkAction.defer, None),
        ('suspicious-header', LinkAction.defer, None),
        # Now if any of the above hit, jump to the hold chain.
        ('any', LinkAction.jump, 'hold'),
        # Finally, the builtin chain jumps to acceptance.
        ('truth', LinkAction.jump, 'accept'),
        )

    def __init__(self):
        self._cached_links = None

    def get_links(self, mlist, msg, msgdata):
        """See `IChain`."""
        if self._cached_links is None:
            self._cached_links = links = []
            for rule, action, chain in self._link_descriptions:
                links.append(Link(rule, action, chain))
        return iter(self._cached_links)



# Copyright (C) 2013-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test database schema migrations"""

import unittest
import alembic.command

from contextlib import suppress
from mailman.config import config
from mailman.database.alembic import alembic_cfg
from mailman.database.factory import LAST_STORM_SCHEMA_VERSION, SchemaManager
from mailman.database.helpers import is_mysql
from mailman.database.model import Model
from mailman.database.types import SAUnicode
from mailman.interfaces.database import DatabaseError
from mailman.testing.layers import ConfigLayer
from sqlalchemy import Column, Integer, MetaData, Table, text
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.schema import Index
from unittest.mock import patch


class TestSchemaManager(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        # Drop the existing model tables.
        Model.metadata.drop_all(config.db.engine)
        # Drop leftover tables (e.g. Alembic & Storm schema versions).
        md = MetaData()
        md.reflect(bind=config.db.engine)
        for table in md.sorted_tables:
            table.drop(config.db.engine)
        self.schema_mgr = SchemaManager(config.db)

    def tearDown(self):
        self._drop_storm_database()
        # Restore a virgin database.
        Model.metadata.create_all(config.db.engine)

    def _table_exists(self, tablename):
        md = MetaData()
        md.reflect(bind=config.db.engine)
        return tablename in md.tables

    def _create_storm_database(self, revision):
        version_table = Table(
            'version', Model.metadata,
            Column('id', Integer, primary_key=True),
            Column('component', SAUnicode),
            Column('version', SAUnicode),
            )
        version_table.create(config.db.engine)

        with config.db.engine.connect() as conn:
            conn.execute(version_table.insert().values(
                component='schema', version=revision))
            # Other Storm specific changes, those SQL statements hopefully
            # work on all DB engines...
            conn.execute(
                text('ALTER TABLE mailinglist ADD COLUMN acceptable_aliases_id INT'))   # noqa: E501
            conn.commit()
        # In case of MySQL, you cannot create/drop indexes on primary keys
        # manually as it is handled automatically by MySQL.
        if not is_mysql(config.db.engine):
            # Don't pollute our main metadata object, create a new one.
            md = MetaData()
            user_table = Model.metadata.tables['user'].tometadata(md)
            (Index('ix_user__user_id', user_table.c._user_id)
             .drop(bind=config.db.engine))
            Index('ix_user_user_id', user_table.c._user_id).create(
                bind=config.db.engine)
        config.db.commit()

    def _drop_storm_database(self):
        """Remove the leftovers from a Storm DB.

        A drop_all() must be issued afterwards.
        """
        if 'version' in Model.metadata.tables:
            version = Model.metadata.tables['version']
            version.drop(config.db.engine, checkfirst=True)
            Model.metadata.remove(version)
        # If it's nonexistent, PostgreSQL raises a ProgrammingError while
        # SQLite raises an OperationalError. Since MySQL automatically handles
        # indexes for primary keys, don't try doing it with that backend.
        if not is_mysql(config.db.engine):
            with suppress(ProgrammingError, OperationalError):
                md = MetaData()
                user_table = Model.metadata.tables['user'].tometadata(md)
                Index('ix_user_user_id', user_table.c._user_id).drop(
                    bind=config.db.engine)
        config.db.commit()

    def test_current_database(self):
        # The database is already at the latest version.
        alembic.command.stamp(alembic_cfg, revision='head', sql=False)
        with patch('alembic.command') as alembic_command:
            self.schema_mgr.setup_database()
            self.assertFalse(alembic_command.stamp.called)
            self.assertFalse(alembic_command.upgrade.called)

    @patch('alembic.command.upgrade')
    def test_initial(self, alembic_command_upgrade):
        # No existing database.
        self.assertFalse(self._table_exists('mailinglist'))
        self.assertFalse(self._table_exists('alembic_version'))
        # For the initial setup of the database, the upgrade command will not
        # be called.  The tables will be created and then the schema stamped
        # at Alembic's latest revision.
        head_rev = self.schema_mgr.setup_database()
        self.assertFalse(alembic_command_upgrade.called)
        self.assertTrue(self._table_exists('mailinglist'))
        self.assertTrue(self._table_exists('alembic_version'))
        # The current Alembic revision is the same as the initial revision.
        md = MetaData()
        md.reflect(bind=config.db.engine)
        with config.db.engine.begin() as conn:
            current_rev = conn.execute(
                md.tables['alembic_version'].select()).scalar()
        self.assertEqual(current_rev, head_rev)

    @patch('alembic.command')
    def test_storm(self, alembic_command):
        # Existing Storm database.
        Model.metadata.create_all(config.db.engine)
        self._create_storm_database(LAST_STORM_SCHEMA_VERSION)
        self.schema_mgr.setup_database()
        self.assertFalse(alembic_command.stamp.called)
        self.assertTrue(alembic_command.upgrade.called)

    @patch('alembic.command')
    def test_old_storm(self, alembic_command):
        # Existing Storm database in an old version.
        Model.metadata.create_all(config.db.engine)
        self._create_storm_database('001')
        self.assertRaises(DatabaseError, self.schema_mgr.setup_database)
        self.assertFalse(alembic_command.stamp.called)
        self.assertFalse(alembic_command.upgrade.called)

    def test_old_db(self):
        # The database is in an old revision, must upgrade.
        alembic.command.stamp(alembic_cfg, 'head')
        md = MetaData()
        md.reflect(bind=config.db.engine)
        config.db.store.execute(md.tables['alembic_version'].delete())
        config.db.store.execute(md.tables['alembic_version'].insert().values(
            version_num='dummyrevision'))
        config.db.commit()
        with patch('alembic.command') as alembic_command:
            self.schema_mgr.setup_database()
            self.assertFalse(alembic_command.stamp.called)
            self.assertTrue(alembic_command.upgrade.called)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interfaces for the request database.

The request database handles events that must be approved by the list
moderators, such as subscription requests and held messages.
"""

from enum import IntEnum
from public import public
from zope.interface import Attribute, Interface


@public
class RequestType(IntEnum):
    held_message = 1
    subscription = 2
    unsubscription = 3


@public
class IListRequests(Interface):
    """Held requests for a specific mailing list."""

    mailing_list = Attribute(
        """The IMailingList for these requests.""")

    count = Attribute(
        """The total number of requests held for the mailing list.""")

    def count_of(request_type):
        """The total number of requests held of the given request type.

        :param request_type: A `RequestType` enum value.
        :return: An integer.
        """

    def hold_request(request_type, key, data=None):
        """Hold some data for moderator approval.

        :param request_type: A `RequestType` enum value.
        :param key: The key piece of request data being held.
        :param data: Additional optional data in the form of a dictionary that
            is associated with the held request.
        :return: A unique id for this held request.
        """

    held_requests = Attribute(
        """An iterator over the held requests.

        Returned items have two attributes:
         * `id` is the held request's unique id;
         * `type` is a `RequestType` enum value.
        """)

    def of_type(request_type):
        """An iterator over the held requests of the given type.

        Returned items have two  attributes:
         * `id` is the held request's unique id;
         * `type` is a `RequestType` enum value.

         Only items with a matching `type' are returned.
         """

    def get_request(request_id, request_type):
        """Get the data associated with the request id, or None.

        :param request_id: The unique id for the request.
        :type request_id: int
        :param request_type: Optional request type that the requested id must
            match, otherwise no match is returned.
        :type request_type: `RequestType`
        :return: A 2-tuple of the key and data originally held, or None if the
            `request_id` is not in the database.
        """

    def delete_request(request_id):
        """Delete the request associated with the id.

        :param request_id: The unique id for the request.
        :raises KeyError: If `request_id` is not in the database.
        """



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The Mail-Archive.com archiver."""

from mailman.config import config
from mailman.config.config import external_configuration
from mailman.interfaces.archiver import ArchivePolicy, IArchiver
from public import public
from urllib.parse import quote, urljoin
from zope.interface import implementer


@public
@implementer(IArchiver)
class MailArchive:
    """Public archiver at the Mail-Archive.com.

    Messages get archived at http://go.mail-archive.com.
    """

    name = 'mail-archive'
    is_enabled = False

    def __init__(self):
        # Read our specific configuration file
        archiver_config = external_configuration(
            config.archiver.mail_archive.configuration)
        self.base_url = archiver_config.get('general', 'base_url')
        self.recipient = archiver_config.get('general', 'recipient')

    def list_url(self, mlist):
        """See `IArchiver`."""
        if mlist.archive_policy is ArchivePolicy.public:
            return urljoin(self.base_url, quote(mlist.posting_address))
        return None

    def permalink(self, mlist, msg):
        """See `IArchiver`."""
        if mlist.archive_policy is not ArchivePolicy.public:
            return None
        # It is the LMTP server's responsibility to ensure that the message has
        # a Message-ID-Hash header.  For backward compatibility, fallback to
        # searching for X-Message-ID-Hash.  If the message has neither, then
        # there's no permalink.
        message_id_hash = msg.get('message-id-hash')
        if message_id_hash is None:
            message_id_hash = msg.get('x-message-id-hash')
        if message_id_hash is None:
            return None
        if isinstance(message_id_hash, bytes):
            message_id_hash = message_id_hash.decode('ascii')
        return urljoin(self.base_url, message_id_hash)

    def archive_message(self, mlist, msg):
        """See `IArchiver`."""
        if mlist.archive_policy is ArchivePolicy.public:
            config.switchboards['out'].enqueue(
                msg,
                listid=mlist.list_id,
                recipients=[self.recipient])
        return None



# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the message store."""

import os
import unittest

from mailman.config import config
from mailman.interfaces.messages import IMessageStore
from mailman.model.message import Message
from mailman.testing.helpers import specialized_message_from_string as mfs
from mailman.testing.layers import ConfigLayer
from mailman.utilities.email import add_message_hash
from zope.component import getUtility


class TestMessageStore(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._store = getUtility(IMessageStore)

    def test_message_id_required(self):
        # The Message-ID header is required in order to add it to the store.
        message = mfs("""\
Subject: An important message

This message is very important.
""")
        self.assertRaises(ValueError, self._store.add, message)

    def test_get_message_by_hash(self):
        # Messages have an X-Message-ID-Hash header, the value of which can be
        # used to look the message up in the message store.
        msg = mfs("""\
Subject: An important message
Message-ID: <ant>

This message is very important.
""")
        add_message_hash(msg)
        self._store.add(msg)
        self.assertEqual(msg['x-message-id-hash'],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')
        found = self._store.get_message_by_hash(
            'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')
        self.assertEqual(found['message-id'], '<ant>')
        self.assertEqual(found['x-message-id-hash'],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')

    def test_can_delete_missing_message(self):
        # Deleting a message which isn't in the store does not raise an
        # exception.
        msg = mfs("""\
Message-ID: <ant>

""")
        self._store.add(msg)
        self.assertEqual(len(list(self._store.messages)), 1)
        self._store.delete_message('missing')
        self.assertEqual(len(list(self._store.messages)), 1)

    def test_can_survive_missing_message_path(self):
        # Deleting a message which is in the store, but which doesn't appear
        # on the file system does not raise an exception, but still removes
        # the message from the store.
        msg = mfs("""\
Message-ID: <ant>

""")
        self._store.add(msg)
        self.assertEqual(len(list(self._store.messages)), 1)
        # We have to use the SQLAlchemy API because the .get_message_by_*()
        # methods return email Message objects, not IMessages.  The former
        # does not give us the path to the object on the file system.
        row = config.db.store.query(Message).filter_by(
            message_id='<ant>').first()
        os.remove(os.path.join(config.MESSAGES_DIR, row.path))
        self._store.delete_message('<ant>')
        self.assertEqual(len(list(self._store.messages)), 0)

    def test_get_message_missing_returns_none_and_deletes(self):
        # Getting a message which is in the store, but which doesn't appear
        # on the file system returns None and removes the message from the
        # store.
        msg = mfs("""\
Message-ID: <ant>

""")
        self._store.add(msg)
        self.assertEqual(len(list(self._store.messages)), 1)
        # We have to use the SQLAlchemy API because the .get_message_by_*()
        # methods return email Message objects, not IMessages.  The former
        # does not give us the path to the object on the file system.
        row = config.db.store.query(Message).filter_by(
            message_id='<ant>').first()
        os.remove(os.path.join(config.MESSAGES_DIR, row.path))
        self.assertIsNone(self._store.get_message_by_id('<ant>'))
        self.assertEqual(len(list(self._store.messages)), 0)

    def test_message_id_hash(self):
        # The new specification calls for a Message-ID-Hash header,
        # specifically without the X- prefix.
        msg = mfs("""\
Message-ID: <ant>

""")
        self._store.add(msg)
        stored_msg = self._store.get_message_by_id('<ant>')
        self.assertEqual(stored_msg['message-id-hash'],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')
        # For backward compatibility with the old spec.
        self.assertEqual(stored_msg['x-message-id-hash'],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')

    def test_message_id_hash_gets_replaced(self):
        # Any existing Message-ID-Hash header (or for backward compatibility
        # X-Message-ID-Hash) gets replaced with its new value.
        msg = mfs("""\
Subject: Testing
Message-ID: <ant>
Message-ID-Hash: abc
X-Message-ID-Hash: abc

""")
        self._store.add(msg)
        stored_msg = self._store.get_message_by_id('<ant>')
        message_id_hashes = stored_msg.get_all('message-id-hash')
        self.assertEqual(len(message_id_hashes), 1)
        self.assertEqual(message_id_hashes[0],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')
        # For backward compatibility with the old spec.
        x_message_id_hashes = stored_msg.get_all('x-message-id-hash')
        self.assertEqual(len(x_message_id_hashes), 1)
        self.assertEqual(x_message_id_hashes[0],
                         'MS6QLWERIJLGCRF44J7USBFDELMNT2BW')

    def test_add_message_duplicate_okay(self):
        msg = mfs("""\
Subject: Once
Message-ID: <ant>

""")
        hash32 = self._store.add(msg)
        stored_msg = self._store.get_message_by_id('<ant>')
        self.assertEqual(msg['subject'], stored_msg['subject'])
        self.assertEqual(msg['message-id-hash'], hash32)
        # A second insertion, even if headers change, does not store the
        # message twice.
        del msg['subject']
        msg['Subject'] = 'Twice'
        hash32 = self._store.add(msg)
        stored_msg = self._store.get_message_by_id('<ant>')
        self.assertNotEqual(msg['subject'], stored_msg['subject'])
        self.assertIsNone(hash32)



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Package and module utilities."""

import os
import sys

from contextlib import contextmanager
from importlib import import_module
from importlib_resources import contents, is_resource, path
from public import public


@public
def abstract_component(cls):
    """Decorator preventing `find_components()` from instantiating the class.

    Normally, `find_components()` instantiates any component class that
    it finds matching the given interface.  Some component classes must not be
    instantiated though, because they act as base classes.  Put this decorator
    on the class definition to prevent instantiation.
    """
    cls.__abstract_component__ = True
    return cls


@public
def find_name(dotted_name):
    """Import and return the named object in package space.

    :param dotted_name: The dotted module path name to the object.
    :type dotted_name: string
    :return: The object.
    :rtype: object
    """
    module_path, dot, object_name = dotted_name.rpartition('.')
    module = import_module(module_path)
    return getattr(module, object_name)


@public
def call_name(dotted_name, *args, **kws):
    """Imports and calls the named object in package space.

    :param dotted_name: The dotted module path name to the object.
    :type dotted_name: string
    :param args: The positional arguments.
    :type args: tuple
    :param kws: The keyword arguments.
    :type kws: dict
    :return: The object.
    :rtype: object
    """
    named_callable = find_name(dotted_name)
    return named_callable(*args, **kws)


@public
def expand_path(resources, url):
    """Expand a python: path, returning the absolute file system path."""
    # Is the context coming from a file system or Python path?
    if url.startswith('python:'):
        resource_path = url[7:]
        package, dot, resource = resource_path.rpartition('.')
        cfg_path = resources.enter_context(path(package, resource + '.cfg'))
        return str(cfg_path)
    else:
        return url


@public
@contextmanager
def hacked_sys_modules(name, module):
    old_module = sys.modules.get(name)
    sys.modules[name] = module
    try:
        yield
    finally:
        if old_module is None:
            del sys.modules[name]
        else:
            sys.modules[name] = old_module


def scan_module(module, interface):
    """Return all the items in a module that conform to an interface.

    Scan every item named in the module's `__all__`.  If that item conforms to
    the given interface, *and* the item is not declared as an
    `@abstract_component`, then return the item.

    :param module: A module object.
    :type module: module
    :param interface: The interface that returned objects must conform to.
    :type interface: `Interface`
    :return: The sequence of matching components.
    :rtype: items implementing `interface`
    """
    missing = object()
    for name in module.__all__:
        component = getattr(module, name, missing)
        assert component is not missing, (
            '%s has bad __all__: %s' % (module, name))   # pragma: nocover
        if (interface.implementedBy(component)
                # We cannot use getattr() here because that will return True
                # for all subclasses.  __abstract_component__ should *not* be
                # inherited, meaning subclasses must declare themselves to be
                # abstract if they also don't want to be instantiated.  Only
                # by looking at the component's __dict__ can we know for sure
                # where the marker has been placed.  The value of
                # __abstract_component__ doesn't matter, only its presence.
                and '__abstract_component__' not in component.__dict__):
            yield component


def find_components(package, interface):
    """Find components which conform to a given interface.

    Search all the modules in a given package, returning an iterator over all
    items found that conform to the given interface, unless that object is
    decorated with `@abstract_component`.

    :param package: The package path to search.
    :type package: string
    :param interface: The interface that returned objects must conform to.
    :type interface: `Interface`
    :return: The sequence of matching components.
    :rtype: items implementing `interface`
    """
    for filename in contents(package):
        basename, extension = os.path.splitext(filename)
        if extension != '.py' or basename.startswith('.'):
            continue
        module_name = '{}.{}'.format(package, basename)
        module = import_module(module_name)
        if not hasattr(module, '__all__'):
            continue
        yield from scan_module(module, interface)


def find_pluggable_components(subpackage, interface):
    """Find components which conform to a given interface.

    This finds components which can be implemented in a plugin.  It will
    search for the interface in the named subpackage, where the Python import
    path of the subpackage will be prepended by `mailman` for system
    components, and the various plugin names for any external components.

    :param subpackage: The subpackage to search.  This is prepended by
        'mailman' to search for system components, and each enabled plugin for
        external components.
    :type subpackage: str
    :param interface: The interface that returned objects must conform to.
    :type interface: `Interface`
    :return: The sequence of matching components.
    :rtype: Objects implementing `interface`
    """
    # This can't be imported at module level because of circular imports.
    from mailman.config import config

    # Return the system components first.
    yield from find_components('mailman.' + subpackage, interface)
    # Return all the matching components in all the subpackages of all enabled
    # plugins.  Only enabled and existing plugins will appear in this
    # dictionary.
    for name, plugin_config in config.plugin_configs:
        # If the plugin's configuration defines a components package, use
        # that, falling back to the plugin's name.
        package = plugin_config['component_package'].strip()
        if len(package) == 0:
            package = name
        # It's possible that the plugin doesn't include the directory for this
        # subpackage.  That's fine.
        if (subpackage in contents(package) and
                not is_resource(package, subpackage)):
            plugin_package = '{}.{}'.format(package, subpackage)
            yield from find_components(plugin_package, interface)


@public
def add_components(subpackage, interface, mapping):
    """Add components to a given mapping.

    Similarly to `find_pluggable_components()` this inspects all modules
    in the given subpackage, relative to the 'mailman' parent package,
    and all the plugin names, that match the given interface.  All such
    found objects (unless decorated with `@abstract_component`) are
    instantiated and added to the given mapping, keyed by the object's `.name`
    attribute, which is required.  It is a fatal error if that key already
    exists in the mapping.

    :param subpackage: The subpackage path to search.
    :type subpackage: str
    :param interface: The interface that returned objects must conform to.
        Objects found must have a `.name` attribute containing a unique
        string.
    :type interface: `Interface`
    :param mapping: The mapping to add the found components to.
    :type mapping: A dict-like mapping.  This only needs to support
        containment tests (e.g. `in` and `not in`) and `__setitem__()`.
    :raises RuntimeError: when a duplicate key is found.
    """
    for component_class in find_pluggable_components(subpackage, interface):
        component = component_class()
        if component.name in mapping:
            raise RuntimeError(     # pragma: nocover
                'Duplicate key "{}" found in {}; previously {}'.format(
                    component.name, component, mapping[component.name]))
        mapping[component.name] = component
